@@ -17,7 +17,8 @@ where
     \;.
 """
 
-from its2D_1 import *
+from __future__ import absolute_import
+from examples.linear_elasticity.its2D_1 import *
 
 from sfepy.mechanics.matcoefs import stiffness_from_youngpoisson
 
