@@ -15,7 +15,7 @@ static void debprintf(const char *what, ...)
 
 int32 mesh_init(Mesh *mesh)
 {
-  int32 ii;
+  uint32 ii;
   MeshTopology *topology = 0;
   LocalEntities *entities = 0;
 
@@ -83,9 +83,9 @@ int32 mesh_free(Mesh *mesh)
 
 int32 mesh_print(Mesh *mesh, FILE *file, int32 header_only)
 {
-  int32 ii, id;
+  uint32 ii, id;
   uint32 *num;
-  int32 D = mesh->topology->max_dim;
+  uint32 D = mesh->topology->max_dim;
   MeshGeometry *geometry = mesh->geometry;
   MeshTopology *topology = mesh->topology;
   MeshConnectivity *conn = 0;
@@ -135,7 +135,7 @@ inline int32 mei_init_conn(MeshEntityIterator *iter, MeshEntity *entity,
                            uint32 dim)
 {
   Mesh *mesh = entity->mesh;
-  int32 D = mesh->topology->max_dim;
+  uint32 D = mesh->topology->max_dim;
   MeshConnectivity *conn = mesh->topology->conn[IJ(D, entity->dim, dim)];
 
   iter->entity->mesh = mesh;
@@ -184,14 +184,16 @@ inline int32 mei_go(MeshEntityIterator *iter)
 inline int32 mei_next(MeshEntityIterator *iter)
 {
   iter->it += 1;
-  iter->entity->ii = (iter->ptr ? iter->ptr[iter->it] : iter->it);
+  if (iter->it < iter->it_end) {
+    iter->entity->ii = (iter->ptr ? iter->ptr[iter->it] : iter->it);
+  }
 
   return(RET_OK);
 }
 
 int32 ind_print(Indices *ind, FILE *file)
 {
-  int32 ii;
+  uint32 ii;
   if (!ind) return(RET_OK);
 
   fprintf(file, "indices: num: %d\n", ind->num);
@@ -271,7 +273,7 @@ int32 conn_free(MeshConnectivity *conn)
 
 int32 conn_print(MeshConnectivity *conn, FILE *file)
 {
-  int32 ii, ic;
+  uint32 ii, ic;
   if (!conn) return(RET_OK);
 
   fprintf(file, "conn: num: %d, n_incident: %d\n", conn->num, conn->n_incident);
@@ -341,7 +343,7 @@ int32 mesh_setup_connectivity(Mesh *mesh, int32 d1, int32 d2)
   int32 ret = RET_OK;
   int32 d3 = 0;
   MeshTopology *topology = mesh->topology;
-  int32 D = topology->max_dim;
+  uint32 D = topology->max_dim;
 
   debprintf("request connectivity %d -> %d\n", d1, d2);
 
@@ -384,7 +386,7 @@ int32 mesh_setup_connectivity(Mesh *mesh, int32 d1, int32 d2)
 
 int32 mesh_free_connectivity(Mesh *mesh, int32 d1, int32 d2)
 {
-  int32 D = mesh->topology->max_dim;
+  uint32 D = mesh->topology->max_dim;
   MeshConnectivity *conn = 0;
 
   debprintf("free connectivity %d -> %d\n", d1, d2);
@@ -400,7 +402,7 @@ int32 mesh_build(Mesh *mesh, int32 dim)
   int32 ret = RET_OK;
   uint32 n_incident, n_v_max, n_loc;
   uint32 ii, ic, id, found;
-  int32 D = mesh->topology->max_dim;
+  uint32 D = mesh->topology->max_dim;
   uint32 facet[4]; // Max. space for single facet.
   uint32 *oris = 0;
   uint32 loc_oris[12];
@@ -450,6 +452,7 @@ int32 mesh_build(Mesh *mesh, int32 dim)
   n_incident = cDd->offsets[cDd->num];
   debprintf("build n_incident (%d -> %d): %d\n", D, dim, n_incident);
 
+  // Cell-local orientations w.r.t. D -> d.
   oris = alloc_mem(uint32, n_incident);
   if (dim == 2) {
     free_mem(mesh->topology->face_oris);
@@ -542,6 +545,7 @@ int32 mesh_build(Mesh *mesh, int32 dim)
         }
         ic--;
       }
+      // printf("%d << %d, %d\n", ic, ii, loc_oris[ii]);
       oris[ic] = loc_oris[ii];
     }
   }
@@ -569,7 +573,7 @@ int32 mesh_transpose(Mesh *mesh, int32 d1, int32 d2)
   uint32 n_incident;
   uint32 ii;
   uint32 *nd2 = 0;
-  int32 D = mesh->topology->max_dim;
+  uint32 D = mesh->topology->max_dim;
   MeshEntityIterator it2[1], it1[1];
   MeshConnectivity *c12 = 0; // d1 -> d2 - to compute
 
@@ -624,7 +628,7 @@ int32 mesh_transpose(Mesh *mesh, int32 d1, int32 d2)
 int32 mesh_intersect(Mesh *mesh, int32 d1, int32 d2, int32 d3)
 {
   int32 ret = RET_OK;
-  int32 D = mesh->topology->max_dim;
+  uint32 D = mesh->topology->max_dim;
   uint32 n_incident, ii;
   uint32 *nd2 = 0;
   char *mask = 0;
@@ -740,7 +744,7 @@ uint32 mesh_count_incident(Mesh *mesh, int32 dim,
 {
   uint32 ii, num = 0;
   uint32 *ptr;
-  int32 D = mesh->topology->max_dim;
+  uint32 D = mesh->topology->max_dim;
   MeshConnectivity *conn = mesh->topology->conn[IJ(D, dent, dim)];
 
   if (!conn->num) {
@@ -764,7 +768,7 @@ int32 mesh_get_incident(Mesh *mesh,
 {
   int32 ret = RET_OK;
   uint32 ii;
-  int32 D = mesh->topology->max_dim;
+  uint32 D = mesh->topology->max_dim;
   MeshEntityIterator it0[1], it1[1];
   MeshConnectivity *conn = mesh->topology->conn[IJ(D, dent, dim)];
 
@@ -793,7 +797,7 @@ int32 mesh_get_local_ids(Mesh *mesh, Indices *local_ids,
 {
   int32 ret = RET_OK;
   uint32 ii, iind, ic, found;
-  int32 D = mesh->topology->max_dim;
+  uint32 D = mesh->topology->max_dim;
   MeshEntity entity[1];
   MeshEntityIterator it1[1];
   MeshConnectivity *conn = mesh->topology->conn[IJ(D, dim, dent)];
@@ -839,7 +843,7 @@ int32 mesh_select_complete(Mesh *mesh, Mask *mask, int32 dim,
                            Indices *entities, int32 dent)
 {
   int32 ret = RET_OK;
-  int32 D = mesh->topology->max_dim;
+  uint32 D = mesh->topology->max_dim;
   uint32 ii, inum;
   char *ent_mask = 0;
   MeshEntityIterator it0[1], it1[1];
@@ -882,7 +886,7 @@ int32 mesh_select_complete(Mesh *mesh, Mask *mask, int32 dim,
 // `ccoors` must be preallocated.
 int32 mesh_get_centroids(Mesh *mesh, float64 *ccoors, int32 dim)
 {
-  int32 nc = mesh->geometry->dim;
+  uint32 nc = mesh->geometry->dim;
   uint32 id;
   float64 *ptr = ccoors;
   float64 *coors = mesh->geometry->coors;
@@ -906,13 +910,173 @@ int32 mesh_get_centroids(Mesh *mesh, float64 *ccoors, int32 dim)
   return(RET_OK);
 }
 
+static inline float64 _det3x3(float64 j[9])
+{
+  return (j[0]*j[4]*j[8] + j[3]*j[7]*j[2] + j[1]*j[5]*j[6]
+          - j[2]*j[4]*j[6] - j[5]*j[7]*j[0] - j[1]*j[3]*j[8]);
+}
+
+static inline float64 _tri_area(float64 *coors, uint32 *indices, uint32 nc)
+{
+#define VS(ic, id) (coors[nc*indices[ic] + id])
+
+  uint32 id;
+  float64 vv0;
+  float64 v0[3], v1[3], ndir[3];
+
+  if (nc == 2) { // 2D.
+    v0[2] = 0.0;
+    v1[2] = 0.0;
+  }
+
+  for (id = 0; id < nc; id++) {
+    vv0 = VS(0, id);
+    v0[id] = VS(1, id) - vv0;
+    v1[id] = VS(2, id) - vv0;
+  }
+  gtr_cross_product(ndir, v0, v1);
+  if (nc == 2) {
+    return 0.5 * fabs(ndir[2]);
+  } else {
+    return 0.5 * sqrt(ndir[0] * ndir[0] + ndir[1] * ndir[1]
+                      + ndir[2] * ndir[2]);
+  }
+
+#undef VS
+}
+
+static inline float64 _aux_hex(float64 *coors, uint32 *indices, uint32 nc,
+                               int32 h, int32 i, int32 j, int32 k)
+{
+#define VS(ic, id) (coors[nc*indices[ic] + id])
+
+  float64 mtx[9];
+
+  mtx[0] = VS(i, 0) + VS(h, 0);
+  mtx[1] = VS(i, 1) + VS(h, 1);
+  mtx[2] = VS(i, 2) + VS(h, 2);
+  mtx[3] = VS(j, 0) - VS(i, 0);
+  mtx[4] = VS(j, 1) - VS(i, 1);
+  mtx[5] = VS(j, 2) - VS(i, 2);
+  mtx[6] = VS(k, 0) - VS(h, 0);
+  mtx[7] = VS(k, 1) - VS(h, 1);
+  mtx[8] = VS(k, 2) - VS(h, 2);
+
+  return _det3x3(mtx);
+
+#undef VS
+}
+
+// `volumes` must be preallocated.
+int32 mesh_get_volumes(Mesh *mesh, float64 *volumes, int32 dim)
+{
+#define VS(ic, id) (coors[nc*entity_vertices->indices[ic] + id])
+
+  int32 ret = RET_OK;
+  uint32 D = mesh->topology->max_dim;
+  uint32 nc = mesh->geometry->dim;
+  uint32 id;
+  uint32 indx2[6];
+  float64 vol, aux, vv0, vv1, vv2, vv3;
+  float64 *ptr = volumes;
+  float64 *coors = mesh->geometry->coors;
+  float64 v0[3], v1[3], v2[3], ndir[3];
+  Indices entity_vertices[1];
+  MeshEntityIterator it0[1];
+  MeshConnectivity *cd0 = 0; // d -> 0
+
+  if (!dim) {
+    errput("vertices have no volume!\n");
+    ERR_CheckGo(ret);
+  }
+
+  cd0 = mesh->topology->conn[IJ(D, dim, 0)];
+
+  for (mei_init(it0, mesh, dim); mei_go(it0); mei_next(it0)) {
+    me_get_incident2(it0->entity, entity_vertices, cd0);
+    /* mei_print(it0, stdout); */
+    /* ind_print(entity_vertices, stdout); */
+    vol = 0;
+
+    if (dim == 1) { // Edges.
+      for (id = 0; id < nc; id++) {
+        aux = VS(1, id) - VS(0, id);
+        vol += aux * aux;
+      }
+      ptr[0] = sqrt(vol);
+
+    } else if (entity_vertices->num == 3) { // Triangles.
+      ptr[0] = _tri_area(coors, entity_vertices->indices, nc);
+
+    } else if (nc == 2) { // Quadrilateral cells.
+      ptr[0] = _tri_area(coors, entity_vertices->indices, nc);
+      indx2[0] = entity_vertices->indices[2];
+      indx2[1] = entity_vertices->indices[3];
+      indx2[2] = entity_vertices->indices[0];
+      ptr[0] += _tri_area(coors, indx2, nc);
+
+    } else if (nc == 3) { // 3D.
+      if (entity_vertices->num == 4) {
+        if (dim == 2) { // Quadrilateral faces (approximate).
+          indx2[0] = entity_vertices->indices[0];
+          indx2[1] = entity_vertices->indices[1];
+          indx2[2] = entity_vertices->indices[2];
+          indx2[3] = entity_vertices->indices[3];
+          indx2[4] = entity_vertices->indices[0];
+          indx2[5] = entity_vertices->indices[1];
+
+          aux = _tri_area(coors, indx2, nc);
+          aux += _tri_area(coors, indx2 + 1, nc);
+          aux += _tri_area(coors, indx2 + 2, nc);
+          aux += _tri_area(coors, indx2 + 3, nc);
+          ptr[0] = 0.5 * aux;
+
+        } else { // Tetrahedral cells.
+          for (id = 0; id < nc; id++) {
+            vv0 = VS(0, id);
+            vv1 = VS(1, id);
+            vv2 = VS(2, id);
+            vv3 = VS(3, id);
+
+            v0[id] = vv1 - vv0;
+            v1[id] = vv2 - vv0;
+            v2[id] = vv3 - vv2;
+          }
+          gtr_cross_product(ndir, v0, v1);
+          gtr_dot_v3(ptr, v2, ndir, 3);
+          ptr[0] /= 6.0;
+        }
+
+      } else { // Hexahedral cells with trilinear interpolation.
+        // See https://math.stackexchange.com/questions/1628540/what-is-the-enclosed-volume-of-an-irregular-cube-given-the-x-y-z-coordinates-of
+        // Uses 0 1 3 2 4 5 7 6 ordering w.r.t. sfepy.
+        aux = _aux_hex(coors, entity_vertices->indices, nc, 0, 1, 3, 2);
+        aux -= _aux_hex(coors, entity_vertices->indices, nc, 4, 5, 7, 6);
+        aux -= _aux_hex(coors, entity_vertices->indices, nc, 0, 1, 4, 5);
+        aux += _aux_hex(coors, entity_vertices->indices, nc, 3, 2, 7, 6);
+        aux += _aux_hex(coors, entity_vertices->indices, nc, 0, 3, 4, 7);
+        aux -= _aux_hex(coors, entity_vertices->indices, nc, 1, 2, 5, 6);
+        ptr[0] = aux / 12.0;
+
+      }
+    }
+
+    ptr += 1;
+  }
+
+ end_label:
+  return(ret);
+
+#undef VS
+}
+
 // `normals` must be preallocated.
 int32 mesh_get_facet_normals(Mesh *mesh, float64 *normals, int32 which)
 {
 #define VS(ic, id) (coors[nc*cell_vertices->indices[ik[ic]] + id])
 
-  int32 D = mesh->topology->max_dim;
-  int32 nc = mesh->geometry->dim;
+  uint32 D = mesh->topology->max_dim;
+  uint32 nc = mesh->geometry->dim;
   int32 dim = D - 1;
   uint32 ii, id, n_loc;
   uint32 *ik;
@@ -999,7 +1163,7 @@ inline int32 me_get_incident(MeshEntity *entity, Indices *out, int32 dim)
 {
   int32 ret = RET_OK;
   Mesh *mesh = entity->mesh;
-  int32 D = mesh->topology->max_dim;
+  uint32 D = mesh->topology->max_dim;
   MeshConnectivity *conn = mesh->topology->conn[IJ(D, entity->dim, dim)];
 
   if (!conn->num) {
@@ -1116,7 +1280,7 @@ inline int32 sort_local_connectivity(MeshConnectivity *loc, uint32 *oris,
 
 inline void uint32_sort234_copy(uint32 *out, uint32 *p, uint32 num)
 {
-  int32 ii;
+  uint32 ii;
   uint32 work;
 
   for (ii = 0; ii < num; ii++) {
