@@ -1,11 +1,14 @@
+from __future__ import absolute_import
 import re
 from copy import copy
 
 import numpy as nm
 
 from sfepy.base.base import assert_, Struct
+import six
+from six.moves import range
 
-_depends = re.compile('r\.([a-zA-Z_0-9.]+)').findall
+_depends = re.compile('r\.([a-zA-Z_\-0-9.]+)').findall
 
 def get_parents(selector):
     """
@@ -22,14 +25,14 @@ def get_dependency_graph(region_defs):
     """
     graph = {}
     name_to_sort_name = {}
-    for sort_name, rdef in region_defs.iteritems():
+    for sort_name, rdef in six.iteritems(region_defs):
         name, sel = rdef.name, rdef.select
-        if name_to_sort_name.has_key(name):
+        if name in name_to_sort_name:
             msg = 'region %s/%s already defined!' % (sort_name, name)
             raise ValueError(msg)
         name_to_sort_name[name] = sort_name
 
-        if not graph.has_key(name):
+        if name not in graph:
             graph[name] = [0]
 
         for parent in get_parents(sel):
@@ -49,7 +52,7 @@ def sort_by_dependency(graph):
     while idone < n_nod:
 
         dep_removed = 0
-        for node, deps in graph.iteritems():
+        for node, deps in six.iteritems(graph):
 
             if (len(deps) == 1) and not deps[0]:
                 out.append(node)
@@ -69,7 +72,7 @@ def sort_by_dependency(graph):
                         dep_removed += 1
 
         if (idone <= idone0) and not dep_removed:
-            raise ValueError, 'circular dependency'
+            raise ValueError('circular dependency')
         idone0 = idone
 
     return out
@@ -241,7 +244,7 @@ class Region(Struct):
                         tdim=tdim, kind_tdim=None,
                         entities=[None] * (tdim + 1),
                         kind=None, parent=parent, shape=None,
-                        mirror_region=None, is_empty=False)
+                        mirror_regions={}, mirror_maps={}, is_empty=False)
         self.set_kind(kind)
 
     def set_kind(self, kind):
@@ -440,13 +443,20 @@ class Region(Struct):
 
         cmesh = self.domain.cmesh
         if idim <= dim:
-            if not allow_lower:
+            if not (allow_lower or allow_empty):
                 msg = 'setup_from_highest() can be used only with dim < %d'
                 raise ValueError(msg % idim)
 
-            cmesh.setup_connectivity(dim, idim)
-            ents = self.get_entities(idim)
-            self.entities[dim] = cmesh.get_complete(dim, ents, idim)
+            if allow_lower:
+                cmesh.setup_connectivity(dim, idim)
+                ents = self.get_entities(idim)
+                self.entities[dim] = cmesh.get_complete(dim, ents, idim)
+
+            else:
+                for idim in range(self.kind_tdim - 1, -1, -1):
+                    self.entities[idim] = nm.empty(0, dtype=nm.uint32)
+
+                self.is_empty = True
 
         else:
             cmesh.setup_connectivity(idim, dim)
@@ -473,14 +483,21 @@ class Region(Struct):
         from the kind entities.
         """
         self._access(self.kind_tdim)
+        is_empty = self.entities[self.kind_tdim].shape[0] == 0
+        if allow_empty:
+            self.is_empty = is_empty
+
+        elif is_empty:
+            raise ValueError('region "%s" has no entities!' % self.name)
+
         for idim in range(self.kind_tdim - 1, -1, -1):
             if self.can[idim] and self.entities[idim] is not None:
                 try:
                     self.setup_from_highest(idim, allow_lower=False,
                                             allow_empty=allow_empty)
 
-                except ValueError, exc:
-                    msg = '\n'.join((exc.message,
+                except ValueError as exc:
+                    msg = '\n'.join((str(exc),
                                      'fix region kind? (region: %s, kind: %s)'
                                      % (self.name, self.kind)))
                     raise ValueError(msg)
@@ -545,7 +562,7 @@ class Region(Struct):
         n_cell = self.cells.shape[0]
         n_edge = self.edges.shape[0] if self.tdim > 1 else 0
         n_face = self.faces.shape[0] if self.tdim == 3 else 0
-        n_facet = self.facets.shape[0] if self.tdim > 1 else 0
+        n_facet = self.facets.shape[0]
 
         self.shape = Struct(n_vertex=n_vertex,
                             n_edge=n_edge,
@@ -617,7 +634,12 @@ class Region(Struct):
         """
         fcells = self.get_cells(true_cells_only=true_cells_only)
 
-        if len(nm.unique(cells)) <= len(nm.unique(fcells)):
+        ucells = nm.unique(cells)
+        ufcells = nm.unique(fcells)
+        if not len(nm.intersect1d(ucells, ufcells, assume_unique=True)):
+            return nm.array([], dtype=nm.int)
+
+        if len(ucells) <= len(ufcells):
             # self is a superset of cells.
             ii = nm.searchsorted(fcells, cells)
             assert_((fcells[ii] == cells).all())
@@ -655,26 +677,73 @@ class Region(Struct):
 
         return fis
 
-    def setup_mirror_region(self):
+    def setup_mirror_region(self, mirror_name=None, ret_name=False):
         """
         Find the corresponding mirror region, set up element mapping.
         """
+        from sfepy.discrete.fem.mesh import find_map
+
         regions = self.domain.regions
+        eopts = self.extra_options
 
-        for reg in regions:
-            mirror_parent = regions.find(reg.parent)
-            if mirror_parent is None: continue
-            if ((reg is not self)
-                and nm.all(self.vertices == reg.vertices)):
-                mirror_region = reg
-                break
+        if (mirror_name is None) and (eopts is not None)\
+            and ('mirror_region' in eopts):
+            mirror_name = eopts['mirror_region']
+
+        if mirror_name is not None:
+            if mirror_name in self.mirror_regions:
+                return
+
+            mreg = regions[mirror_name]
+            if self.vertices.shape[0] != mreg.vertices.shape[0]:
+                raise ValueError('%s: incompatible mirror region! (%s)'
+                    % (self.name, mreg.name))
+            coors = self.domain.cmesh.coors
+            coors1 = coors[self.vertices, :]
+            coors2 = coors[mreg.vertices, :]
+            shift = ((nm.sum(coors2, axis=0) - nm.sum(coors1, axis=0))
+                / coors1.shape[0])
+            vmap1, vmap2 = find_map(coors1, coors2 - shift, join=False)
+            if vmap1.shape[0] != coors1.shape[0]:
+                print(coors1[vmap1])
+                print(coors2[vmap2])
+                raise ValueError('cannot match vertices!')
+
+            fconn = self.domain.cmesh.get_conn_as_graph(self.dim - 1, 0)
+            cc = self.domain.cmesh.get_centroids(self.dim - 1)
+            fmap1, fmap2 = find_map(cc[self.facets], cc[mreg.facets] - shift,
+                                    join=False)
+            if fmap1.shape[0] != self.facets.shape[0]:
+                print(cc[self.facets][fmap1])
+                print(cc[mreg.facets][fmap2])
+                raise ValueError('cannot match facets!')
+
+            mirror_map_i = nm.zeros_like(fmap1)
+            mirror_map_i[fmap1] = fmap2
+            mirror_map = nm.zeros_like(fmap1)
+            mirror_map[fmap2] = fmap1
         else:
-            raise ValueError('cannot find mirror region! (%s)' % self.name)
+            for reg in regions:
+                mirror_parent = regions.find(reg.parent)
+                if mirror_parent is None: continue
+                if ((reg is not self)
+                    and nm.all(self.vertices == reg.vertices)):
+                    mreg = reg
+                    mirror_map = mirror_map_i = None
+                    break
+            else:
+                raise ValueError('cannot find mirror region! (%s)' % self.name)
 
-        self.mirror_region = mirror_region
+        self.mirror_regions[mreg.name] = mreg
+        self.mirror_maps[mreg.name] = mirror_map
+        mreg.mirror_regions[self.name] = self
+        mreg.mirror_maps[self.name] = mirror_map_i
 
-    def get_mirror_region(self):
-        return self.mirror_region
+        if ret_name:
+            return mreg.name
+
+    def get_mirror_region(self, name):
+        return self.mirror_regions[name]
 
     def get_n_cells(self, is_surface=False):
         """
@@ -748,7 +817,7 @@ class Region(Struct):
         cmesh = self.domain.cmesh
 
         e_verts = cmesh.get_incident(0, self.edges, 1)
-        e_verts.shape = (e_verts.shape[0] / 2, 2)
+        e_verts.shape = (e_verts.shape[0] // 2, 2)
 
         ii = nm.where(e_verts[:, 0] != e_verts[:, 1])[0]
         edges = self.edges[ii]
