@@ -1,7 +1,7 @@
 """
 Classes of equations composed of terms.
 """
-import time
+from __future__ import absolute_import
 from copy import copy
 
 import numpy as nm
@@ -9,15 +9,18 @@ import scipy.sparse as sp
 
 from sfepy.base.base import output, assert_, get_default, iter_dict_of_lists
 from sfepy.base.base import OneTypeList, Container, Struct
+from sfepy.base.timing import Timer
 from sfepy.discrete import Materials, Variables, create_adof_conns
 from sfepy.discrete.common.extmods.cmesh import create_mesh_graph
 from sfepy.terms import Terms, Term
+from sfepy.terms.terms_multilinear import ETermBase
+import six
 
 def parse_definition(equation_def):
     """
     Parse equation definition string to create term description list.
     """
-    from parse_equations import create_bnf
+    from .parse_equations import create_bnf
 
     term_descs = []
     bnf = create_bnf(term_descs)
@@ -49,19 +52,20 @@ class Equations(Container):
 
     @staticmethod
     def from_conf(conf, variables, regions, materials, integrals,
-                  user=None, verbose=True):
+                  user=None, eterm_options=None, verbose=True):
 
         objs = OneTypeList(Equation)
 
         conf = copy(conf)
 
         ii = 0
-        for name, desc in conf.iteritems():
+        for name, desc in six.iteritems(conf):
             if verbose:
                 output('equation "%s":' %  name)
                 output(desc)
             eq = Equation.from_desc(name, desc, variables, regions,
-                                    materials, integrals, user=user)
+                                    materials, integrals, user=user,
+                                    eterm_options=eterm_options)
             objs.append(eq)
             ii += 1
 
@@ -81,6 +85,26 @@ class Equations(Container):
 
         self.collect_conn_info()
 
+    def add_equation(self, equation):
+        """
+        Add a new equation.
+
+        Parameters
+        ----------
+        equation : Equation instance
+                    The new equation.
+        """
+        self.append(equation)
+        self.variables.extend(
+             set(equation.collect_variables() ) - set(self.variables)
+        )
+        self.materials.extend(
+             set(equation.collect_materials() ) - set(self.materials)
+        )
+        equation.collect_conn_info(self.conn_info)
+        if not self.domain:
+           self.domain = self.get_domain()
+
     def create_subequations(self, var_names, known_var_names=None):
         """
         Create sub-equations containing only terms with the given virtual
@@ -235,9 +259,7 @@ class Equations(Container):
         for eq in self:
             output('  %s:' % eq.name)
             for term in eq.terms:
-                output('    %+.2e * %s.%d.%s(%s)'
-                       % (term.sign, term.name, term.integral.order,
-                          term.region.name, term.arg_str))
+                output('    %s' % term.get_str())
 
     def time_update(self, ts, ebcs=None, epbcs=None, lcbcs=None,
                     functions=None, problem=None, active_only=True,
@@ -283,7 +305,7 @@ class Equations(Container):
         active_bcs = self.variables.equation_mapping(ebcs, epbcs, ts, functions,
                                                      problem=problem,
                                                      active_only=active_only)
-        graph_changed = active_only and (active_bcs != self.active_bcs)
+        graph_changed = active_bcs != self.active_bcs
         self.active_bcs = active_bcs
 
         if graph_changed or not self.variables.adof_conns:
@@ -443,7 +465,8 @@ class Equations(Container):
         shape = get_default(shape, self.variables.get_matrix_shape())
 
         output('matrix shape:', shape, verbose=verbose)
-        if nm.prod(shape) == 0:
+        size = nm.prod(shape, dtype=nm.int64)
+        if size == 0:
             output('no matrix (zero size)!')
             return None
 
@@ -456,14 +479,14 @@ class Equations(Container):
             return None
 
         output('assembling matrix graph...', verbose=verbose)
-        tt = time.clock()
+        timer = Timer(start=True)
 
         nnz, prow, icol = create_mesh_graph(shape[0], shape[1],
                                             len(rdcs), rdcs, cdcs)
 
-        output('...done in %.2f s' % (time.clock() - tt), verbose=verbose)
+        output('...done in %.2f s' % timer.stop(), verbose=verbose)
         output('matrix structural nonzeros: %d (%.2e%% fill)' \
-               % (nnz, float(nnz) / nm.prod(shape)), verbose=verbose)
+               % (nnz, 100.0 * float(nnz) / size), verbose=verbose)
 
         data = nm.zeros((nnz,), dtype=self.variables.dtype)
         matrix = sp.csr_matrix((data, icol, prow), shape)
@@ -608,11 +631,15 @@ class Equations(Container):
             eqs = [self[eq] for eq in names]
 
         if mode == 'weak':
+            extras = []
             for eq in eqs:
-                eq.evaluate(mode=mode, dw_mode=dw_mode,
-                            term_mode=term_mode, asm_obj=asm_obj)
+                out = eq.evaluate(mode=mode, dw_mode=dw_mode,
+                                  term_mode=term_mode, asm_obj=asm_obj)
+                if isinstance(out, tuple): extras.extend(out[1])
 
             out = asm_obj
+            for extra in extras:
+                out = out + extra
 
         else:
             out = {}
@@ -723,17 +750,16 @@ class Equations(Container):
                 ic = get_indx(cname, stripped=True, allow_dual=True)
 
                 tangent_matrix.data[:] = 0.0
-                eq.evaluate(mode='weak', dw_mode='matrix',
-                            asm_obj=tangent_matrix)
+                aux = eq.evaluate(mode='weak', dw_mode='matrix',
+                                  asm_obj=tangent_matrix)
 
-                out[key] = tangent_matrix[ir, ic]
+                out[key] = aux[ir, ic]
 
         else:
             tangent_matrix.data[:] = 0.0
 
-            self.evaluate(mode='weak', dw_mode='matrix', asm_obj=tangent_matrix)
-
-            out = tangent_matrix
+            out = self.evaluate(mode='weak', dw_mode='matrix',
+                                asm_obj=tangent_matrix)
 
         return out
 
@@ -741,13 +767,19 @@ class Equation(Struct):
 
     @staticmethod
     def from_desc(name, desc, variables, regions, materials, integrals,
-                  user=None):
+                  user=None, eterm_options=None):
         term_descs = parse_definition(desc)
         terms = Terms.from_desc(term_descs, regions, integrals)
 
         terms.setup()
         terms.assign_args(variables, materials, user)
 
+        if eterm_options is not None:
+            for term in terms:
+                if isinstance(term, ETermBase):
+                    term.set_verbosity(eterm_options.get('verbosity', 0))
+                    term.set_backend(**eterm_options.get('backend_args', {}))
+
         obj = Equation(name, terms)
 
         return obj
@@ -830,8 +862,11 @@ class Equation(Struct):
                                                       ret_status=True)
                     term.assemble_to(asm_obj, val, iels, mode=dw_mode)
 
+                out = asm_obj
+
             elif dw_mode == 'matrix':
 
+                extras = []
                 for term in self.terms:
                     svars = term.get_state_variables(unknown_only=True)
 
@@ -841,14 +876,15 @@ class Equation(Struct):
                                                           diff_var=svar.name,
                                                           standalone=False,
                                                           ret_status=True)
-                        term.assemble_to(asm_obj, val, iels,
-                                         mode=dw_mode, diff_var=svar)
+                        extra = term.assemble_to(asm_obj, val, iels,
+                                                 mode=dw_mode, diff_var=svar)
+                        if extra is not None: extras.append(extra)
+
+                out = (asm_obj, extras) if len(extras) else asm_obj
 
             else:
                 raise ValueError('unknown assembling mode! (%s)' % dw_mode)
 
-            out = asm_obj
-
         else:
             raise ValueError('unknown evaluation mode! (%s)' % mode)
 
