@@ -11,32 +11,37 @@ Important attributes of continuous (order > 0) :class:`Field` and
 where `conn` is the mesh vertex connectivity, `econn` is the
 region-local field connectivity.
 """
+from __future__ import absolute_import
 import numpy as nm
 
 from sfepy.base.base import output, get_default, assert_
 from sfepy.base.base import Struct
+from sfepy.base.timing import Timer
 from sfepy.discrete.common.fields import parse_shape, Field
+from sfepy.discrete import PolySpace
 from sfepy.discrete.fem.mesh import Mesh
 from sfepy.discrete.fem.meshio import convert_complex_output
 from sfepy.discrete.fem.utils import (extend_cell_data, prepare_remap,
                                       invert_remap, get_min_value)
 from sfepy.discrete.fem.mappings import VolumeMapping, SurfaceMapping
-from sfepy.discrete.fem.poly_spaces import PolySpace
 from sfepy.discrete.fem.fe_surface import FESurface
 from sfepy.discrete.integrals import Integral
 from sfepy.discrete.fem.linearizer import (get_eval_dofs, get_eval_coors,
                                            create_output)
+import six
 
 def set_mesh_coors(domain, fields, coors, update_fields=False, actual=False,
-                   clear_all=True):
+                   clear_all=True, extra_dofs=False):
     if actual:
-        domain.mesh.coors_act = coors.copy()
+        if not hasattr(domain.mesh, 'coors_act'):
+            domain.mesh.coors_act = nm.zeros_like(domain.mesh.coors)
+        domain.mesh.coors_act[:] = coors[:domain.mesh.n_nod]
     else:
-        domain.cmesh.coors[:] = coors
+        domain.cmesh.coors[:] = coors[:domain.mesh.n_nod]
 
     if update_fields:
-        for field in fields.itervalues():
-            field.setup_coors(coors)
+        for field in six.itervalues(fields):
+            field.set_coors(coors, extra_dofs=extra_dofs)
             field.clear_mappings(clear_all=clear_all)
 
 def eval_nodal_coors(coors, mesh_coors, region, poly_space, geom_poly_space,
@@ -252,6 +257,10 @@ class FEField(Field):
         self.setup_coors()
         self.clear_mappings(clear_all=True)
         self.clear_qp_base()
+        self.basis_transform = None
+        self.econn0 = None
+        self.unused_dofs = None
+        self.stored_subs = None
 
     def _set_approx_order(self, approx_order):
         """
@@ -347,16 +356,20 @@ class FEField(Field):
                 efs = efs[:,nm.newaxis]
             self.efaces = nm.hstack((self.efaces, efs))
 
-    def setup_coors(self, coors=None):
-        """
-        Setup coordinates of field nodes.
-        """
-        mesh = self.domain.mesh
-        self.coors = nm.empty((self.n_nod, mesh.dim), nm.float64)
+        if gel.dim == 3:
+            self.eedges = gel.edges.copy()
+            efs = node_desc.edge
+            if efs is not None:
+                efs = nm.array(efs).squeeze()
 
-        if coors is None:
-            coors = mesh.coors
+                if efs.ndim < 2:
+                    efs = efs[:,nm.newaxis]
+                self.eedges = nm.hstack((self.eedges, efs))
 
+    def set_coors(self, coors, extra_dofs=False):
+        """
+        Set coordinates of field nodes.
+        """
         # Mesh vertex nodes.
         if self.n_vertex_dof:
             indx = self.vertex_remap_i
@@ -364,10 +377,27 @@ class FEField(Field):
                                                      indx.astype(nm.int32),
                                                      axis=0)
 
-        gps = self.gel.poly_space
-        ps = self.poly_space
+        n_ex_dof = self.n_bubble_dof + self.n_edge_dof + self.n_face_dof
 
-        eval_nodal_coors(self.coors, coors, self.region, ps, gps, self.econn)
+        # extra nodes
+        if n_ex_dof:
+            if extra_dofs:
+                if self.n_nod != coors.shape[0]:
+                    raise NotImplementedError
+                self.coors[:] = coors
+            else:
+                gps = self.gel.poly_space
+                ps = self.poly_space
+                eval_nodal_coors(self.coors, coors, self.region,
+                                 ps, gps, self.econn)
+
+    def setup_coors(self):
+        """
+        Setup coordinates of field nodes.
+        """
+        mesh = self.domain.mesh
+        self.coors = nm.empty((self.n_nod, mesh.dim), nm.float64)
+        self.set_coors(mesh.coors)
 
     def get_vertices(self):
         """
@@ -413,6 +443,11 @@ class FEField(Field):
         dim = region.dim
 
         if integration in ('surface', 'surface_extra'):
+            if region_name not in self.surface_data:
+                reg = self.domain.regions[region_name]
+                self.domain.create_surface_group(reg)
+                self.setup_surface_data(reg, False, None)
+
             sd = self.surface_data[region_name]
 
             # This works also for surface fields.
@@ -496,7 +531,7 @@ class FEField(Field):
         """
         qpkey = (integral.order, key)
 
-        if not self.qp_coors.has_key(qpkey):
+        if qpkey not in self.qp_coors:
             if (key[0] == 's') and not self.is_surface:
                 dim = self.gel.dim - 1
                 n_fp = self.gel.surface_facet.n_vertex
@@ -510,6 +545,83 @@ class FEField(Field):
 
         return self.qp_coors[qpkey]
 
+    def substitute_dofs(self, subs, restore=False):
+        """
+        Perform facet DOF substitutions according to `subs`.
+
+        Modifies `self.econn` in-place and sets `self.econn0`,
+        `self.unused_dofs` and `self.basis_transform`.
+        """
+        if restore and (self.stored_subs is not None):
+            self.econn0 = self.econn
+            self.econn, self.unused_dofs, basis_transform = self.stored_subs
+
+        else:
+            if subs is None:
+                self.econn0 = self.econn
+                return
+
+            else:
+                self.econn0 = self.econn.copy()
+
+            self._substitute_dofs(subs)
+
+            self.unused_dofs = nm.setdiff1d(self.econn0, self.econn)
+
+            basis_transform = self._eval_basis_transform(subs)
+
+        self.set_basis_transform(basis_transform)
+
+    def restore_dofs(self, store=False):
+        """
+        Undoes the effect of :func:`FEField.substitute_dofs()`.
+        """
+        if self.econn0 is None:
+            raise ValueError('no original DOFs to restore!')
+
+        if store:
+            self.stored_subs = (self.econn,
+                                self.unused_dofs,
+                                self.basis_transform)
+
+        else:
+            self.stored_subs = None
+
+        self.econn = self.econn0
+        self.econn0 = None
+        self.unused_dofs = None
+        self.basis_transform = None
+
+    def set_basis_transform(self, transform):
+        """
+        Set local element basis transformation.
+
+        The basis transformation is applied in :func:`FEField.get_base()` and
+        :func:`FEField.create_mapping()`.
+
+        Parameters
+        ----------
+        transform : array, shape `(n_cell, n_ep, n_ep)`
+            The array with `(n_ep, n_ep)` transformation matrices for each cell
+            in the field's region, where `n_ep` is the number of element DOFs.
+        """
+        self.basis_transform = transform
+
+    def restore_substituted(self, vec):
+        """
+        Restore values of the unused DOFs using the transpose of the applied
+        basis transformation.
+        """
+        if (self.econn0 is None) or (self.basis_transform is None):
+            raise ValueError('no original DOF values to restore!!')
+
+        vec = vec.reshape((self.n_nod, self.n_components)).copy()
+        evec = vec[self.econn]
+
+        vec[self.econn0] = nm.einsum('cji,cjk->cik', self.basis_transform, evec)
+
+        return vec.ravel()
+
     def get_base(self, key, derivative, integral, iels=None,
                  from_geometry=False, base_only=True):
         qp = self.get_qp(key, integral)
@@ -523,19 +635,20 @@ class FEField(Field):
         _key = key if not from_geometry else 'g' + key
         bf_key = (integral.order, _key, derivative)
 
-        if not self.bf.has_key(bf_key):
-            if (iels is not None) and (self.ori is not None):
-                ori = self.ori[iels]
+        if bf_key not in self.bf:
+            ori = self.ori
+            self.bf[bf_key] = ps.eval_base(qp.vals, diff=derivative, ori=ori,
+                                           transform=self.basis_transform)
 
-            else:
-                ori = self.ori
-
-            self.bf[bf_key] = ps.eval_base(qp.vals, diff=derivative, ori=ori)
+        bf = self.bf[bf_key]
+        if iels is not None and bf.ndim == 4:
+            bf = bf[iels]
 
         if base_only:
-            return self.bf[bf_key]
+            return bf
+
         else:
-            return self.bf[bf_key], qp.weights
+            return bf, qp.weights
 
     def create_bqp(self, region_name, integral):
         gel = self.gel
@@ -778,8 +891,6 @@ class FEField(Field):
         cache : Struct instance
             The evaluate cache.
         """
-        import time
-
         try:
             from scipy.spatial import cKDTree as KDTree
         except ImportError:
@@ -790,7 +901,7 @@ class FEField(Field):
         if cache is None:
             cache = Struct(name='evaluate_cache')
 
-        tt = time.clock()
+        timer = Timer(start=True)
         if (cache.get('cmesh', None) is None) or not share_geometry:
             mesh = self.create_mesh(extra_nodes=False)
             cache.cmesh = cmesh = mesh.cmesh
@@ -810,13 +921,13 @@ class FEField(Field):
                 cache.normals0 = cmesh.get_facet_normals(0)
                 cache.normals1 = cmesh.get_facet_normals(1)
 
-        output('cmesh setup: %f s' % (time.clock()-tt), verbose=verbose)
+        output('cmesh setup: %f s' % timer.stop(), verbose=verbose)
 
-        tt = time.clock()
+        timer.start()
         if (cache.get('kdtree', None) is None) or not share_geometry:
             cache.kdtree = KDTree(cmesh.coors)
 
-        output('kdtree: %f s' % (time.clock()-tt), verbose=verbose)
+        output('kdtree: %f s' % timer.stop(), verbose=verbose)
 
         return cache
 
@@ -889,11 +1000,14 @@ class FEField(Field):
         coors = domain.get_mesh_coors(actual=True)
         dconn = domain.get_conn()
 
+        tco = integration in ('volume', 'custom')
+        iels = region.get_cells(true_cells_only=tco)
+        transform = (self.basis_transform[iels] if self.basis_transform
+                     is not None else None)
+
         if integration == 'volume':
             qp = self.get_qp('v', integral)
 
-            iels = region.get_cells()
-
             geo_ps = self.gel.poly_space
             ps = self.poly_space
             bf = self.get_base('v', 0, integral, iels=iels)
@@ -901,7 +1015,7 @@ class FEField(Field):
             conn = nm.take(dconn, iels.astype(nm.int32), axis=0)
             mapping = VolumeMapping(coors, conn, poly_space=geo_ps)
             vg = mapping.get_mapping(qp.vals, qp.weights, poly_space=ps,
-                                     ori=self.ori)
+                                     ori=self.ori, transform=transform)
 
             out = vg
 
@@ -913,6 +1027,11 @@ class FEField(Field):
                       ' hierarchical basis!'
                 raise ValueError(msg)
 
+            if self.basis_transform is not None:
+                msg = 'surface integrals do not work with the' \
+                      ' basis transform!'
+                raise ValueError(msg)
+
             sd = domain.surface_groups[region.name]
             esd = self.surface_data[region.name]
 
@@ -926,7 +1045,7 @@ class FEField(Field):
                 self.create_bqp(region.name, integral)
                 qp = self.qp_coors[(integral.order, esd.bkey)]
 
-                abf = ps.eval_base(qp.vals[0])
+                abf = ps.eval_base(qp.vals[0], transform=transform)
                 bf = abf[..., self.efaces[0]]
 
                 indx = self.gel.get_surface_entities()[0]
@@ -947,9 +1066,14 @@ class FEField(Field):
                     sg.evaluate_bfbgm(bf_bg, ebf_bg, coors, sd.fis, dconn)
 
             else:
+                if self.basis_transform is not None:
+                    msg = 'surface fields do not work with the' \
+                          ' basis transform!'
+                    raise ValueError(msg)
+
                 # Do not use BQP for surface fields.
                 qp = self.get_qp(sd.face_type, integral)
-                bf = ps.eval_base(qp.vals)
+                bf = ps.eval_base(qp.vals, transform=transform)
 
                 sg = mapping.get_mapping(qp.vals, qp.weights,
                                          poly_space=Struct(n_nod=bf.shape[-1]),
@@ -1015,7 +1139,7 @@ class VolumeField(FEField):
         Setup the field region geometry.
         """
         cmesh = self.domain.cmesh
-        for key, gel in self.domain.geom_els.iteritems():
+        for key, gel in six.iteritems(self.domain.geom_els):
             ct = cmesh.cell_types
             if (ct[self.region.cells] == cmesh.key_to_index[gel.name]).all():
                 self.gel = gel
@@ -1083,8 +1207,9 @@ class VolumeField(FEField):
 
         if (dct == 'surface') or (geometry_flag):
             reg = info.get_region()
+            mreg_name = info.get_region_name(can_trace=False)
             self.domain.create_surface_group(reg)
-            self.setup_surface_data(reg, is_trace)
+            self.setup_surface_data(reg, is_trace, mreg_name)
 
         elif dct == 'edge':
             raise NotImplementedError('dof connectivity type %s' % dct)
@@ -1101,7 +1226,7 @@ class VolumeField(FEField):
             conn.shape += (1,)
             self.point_data[region.name] = conn
 
-    def setup_surface_data(self, region, is_trace=False):
+    def setup_surface_data(self, region, is_trace=False, trace_region=None):
         """nodes[leconn] == econn"""
         """nodes are sorted by node number -> same order as region.vertices"""
         if region.name not in self.surface_data:
@@ -1111,11 +1236,12 @@ class VolumeField(FEField):
 
         if region.name in self.surface_data and is_trace:
             sd = self.surface_data[region.name]
-            sd.setup_mirror_connectivity(region)
+            sd.setup_mirror_connectivity(region, trace_region)
 
         return self.surface_data[region.name]
 
-    def get_econn(self, conn_type, region, is_trace=False, integration=None):
+    def get_econn(self, conn_type, region, is_trace=False, integration=None,
+                  local=False):
         """
         Get extended connectivity of the given type in the given region.
         """
@@ -1132,8 +1258,12 @@ class VolumeField(FEField):
                 conn = nm.take(self.econn, ii, axis=0)
 
         elif ct == 'surface':
+            if region.name not in self.surface_data:
+                self.domain.create_surface_group(region)
+                self.setup_surface_data(region)
+
             sd = self.surface_data[region.name]
-            conn = sd.get_connectivity(is_trace=is_trace)
+            conn = sd.get_connectivity(local=local, is_trace=is_trace)
 
         elif ct == 'edge':
             raise NotImplementedError('connectivity type %s' % ct)
@@ -1206,16 +1336,24 @@ class SurfaceField(FEField):
         ok : bool
             True if the region is usable for the field.
         """
-        ok = ((region.kind_tdim == (region.tdim - 1))
-              and (region.get_n_cells(True) > 0))
+        ok1 = ((region.kind_tdim == (region.tdim - 1))
+               and (region.get_n_cells(True) > 0))
+        if not ok1:
+            output('bad region topological dimension and kind! (%d, %s)'
+                   % (region.tdim, region.kind))
 
-        return ok
+        n_ns = region.get_facet_indices().shape[0] - region.get_n_cells(True)
+        ok2 = n_ns == 0
+        if not ok2:
+            output('%d region facets are not on the domain surface!' % n_ns)
+
+        return ok1 and ok2
 
     def _setup_geometry(self):
         """
         Setup the field region geometry.
         """
-        for key, vgel in self.domain.geom_els.iteritems():
+        for key, vgel in six.iteritems(self.domain.geom_els):
             self.gel = vgel.surface_facet
             break
 
@@ -1249,7 +1387,8 @@ class SurfaceField(FEField):
 
         if reg.name in self.surface_data and is_trace:
             sd = self.surface_data[reg.name]
-            sd.setup_mirror_connectivity(reg)
+            mreg_name = info.get_region_name(can_trace=False)
+            sd.setup_mirror_connectivity(reg, mreg_name)
 
     def _init_econn(self):
         """
