@@ -1,8 +1,12 @@
+from __future__ import absolute_import
 import numpy as nm
 import numpy.linalg as nla
 
-from sfepy.base.base import find_subclasses, assert_, Struct
+from sfepy.base.base import assert_, Struct
+from sfepy.discrete import PolySpace
 from sfepy.linalg import combine, insert_strided_axis
+from six.moves import range
+from functools import reduce
 
 # Requires fixed vertex numbering!
 vertex_maps = {3 : [[0, 0, 0],
@@ -18,7 +22,8 @@ vertex_maps = {3 : [[0, 0, 0],
                     [1, 1],
                     [0, 1]],
                1 : [[0],
-                    [1]]}
+                    [1]],
+               0 : [[0]]}
 
 class LagrangeNodes(Struct):
     """Helper class for defining nodes of Lagrange elements."""
@@ -99,7 +104,7 @@ class LagrangeNodes(Struct):
     @staticmethod
     def append_tp_faces(nodes, nts, iseq, nt, faces, ao):
         delta = 1.0 / (float(ao) ** 2)
-        for ii, face in enumerate( faces ):
+        for ii, face in enumerate(faces):
             n1 = nodes[face[0],:].copy()
             n2 = nodes[face[1],:].copy()
             n3 = nodes[face[2],:].copy()
@@ -220,138 +225,17 @@ class NodeDescription(Struct):
         """
         return (self.n_edge_nod + self.n_face_nod + self.n_bubble_nod) > 0
 
-class PolySpace(Struct):
-    """Abstract polynomial space class."""
-    _all = None
-
-    keys = {
-        (1, 2) : 'simplex',
-        (2, 3) : 'simplex',
-        (3, 4) : 'simplex',
-        (2, 4) : 'tensor_product',
-        (3, 8) : 'tensor_product',
-    }
-
-    @staticmethod
-    def any_from_args(name, geometry, order, base='lagrange',
-                      force_bubble=False):
-        """
-        Construct a particular polynomial space classes according to the
-        arguments passed in.
-        """
-        if name is None:
-            name = PolySpace.suggest_name(geometry, order, base, force_bubble)
-
-        if PolySpace._all is None:
-            PolySpace._all = find_subclasses(globals(), [PolySpace])
-        table = PolySpace._all
-
-        key = '%s_%s' % (base, PolySpace.keys[(geometry.dim,
-                                               geometry.n_vertex)])
-        if (geometry.name == '1_2') and (key not in table):
-            key = '%s_%s' % (base, 'tensor_product')
-
-        if force_bubble:
-            key += '_bubble'
-
-        return table[key](name, geometry, order)
-
-    @staticmethod
-    def suggest_name(geometry, order, base='lagrange',
-                     force_bubble=False):
-        """
-        Suggest the polynomial space name given its constructor parameters.
-        """
-        aux = geometry.get_interpolation_name()[:-1]
-        if force_bubble:
-            return aux + ('%dB' % order)
-        else:
-            return aux + ('%d' % order)
-
-    def __init__(self, name, geometry, order):
-        self.name = name
-        self.geometry = geometry
-        self.order = order
-
-        self.bbox = nm.vstack((geometry.coors.min(0), geometry.coors.max(0)))
-
-    def eval_base(self, coors, diff=False, ori=None, force_axis=False,
-                  suppress_errors=False, eps=1e-15):
-        """
-        Evaluate the basis in points given by coordinates. The real work is
-        done in _eval_base() implemented in subclasses.
-
-        Parameters
-        ----------
-        coors : array_like
-            The coordinates of points where the basis is evaluated. See Notes.
-        diff : bool
-            If True, return the first derivative.
-        ori : array_like, optional
-            Optional orientation of element facets for per element basis.
-        force_axis : bool
-            If True, force the resulting array shape to have one more axis even
-            when `ori` is None.
-        suppress_errors : bool
-            If True, do not report points outside the reference domain.
-        eps : float
-            Accuracy for comparing coordinates.
-
-        Returns
-        -------
-        base : array
-            The basis (shape (n_coor, 1, n_base)) or its derivative (shape
-            (n_coor, dim, n_base)) evaluated in the given points. An additional
-            axis is pre-pended of length n_cell, if `ori` is given, or of
-            length 1, if `force_axis` is True.
-
-        Notes
-        -----
-        If coors.ndim == 3, several point sets are assumed, with equal number
-        of points in each of them. This is the case, for example, of the
-        values of the volume base functions on the element facets. The indexing
-        (of bf_b(g)) is then (ifa,iqp,:,n_ep), so that the facet can be set in
-        C using FMF_SetCell.
-        """
-        coors = nm.asarray(coors)
-        if not coors.ndim in (2, 3):
-            raise ValueError('coordinates must have 2 or 3 dimensions! (%d)'
-                             % coors.ndim)
-
-        if (coors.ndim == 2):
-            base = self._eval_base(coors, diff=diff, ori=ori,
-                                   suppress_errors=suppress_errors,
-                                   eps=eps)
-
-            if (base.ndim == 3) and force_axis:
-                base = base[None, ...]
-
-            if not base.flags['C_CONTIGUOUS']:
-                base = nm.ascontiguousarray(base)
-
-        else: # Several point sets.
-            if diff:
-                bdim = self.geometry.dim
-            else:
-                bdim = 1
-
-            base = nm.empty((coors.shape[0], coors.shape[1],
-                             bdim, self.n_nod), dtype=nm.float64)
-
-            for ii, _coors in enumerate(coors):
-                base[ii] = self._eval_base(_coors, diff=diff, ori=ori,
-                                           suppress_errors=suppress_errors,
-                                           eps=eps)
-
-        return base
-
+class FEPolySpace(PolySpace):
+    """
+    Base for FE polynomial space classes.
+    """
     def get_mtx_i(self):
         return self.mtx_i
 
     def describe_nodes(self):
         return NodeDescription(self.nts, self.nodes)
 
-class LagrangePolySpace(PolySpace):
+class LagrangePolySpace(FEPolySpace):
 
     def create_context(self, cmesh, eps, check_errors, i_max, newton_eps,
                        tdim=None):
@@ -388,14 +272,18 @@ class LagrangePolySpace(PolySpace):
 
         return ctx
 
-    def _eval_base(self, coors, diff=False, ori=None,
+    def _eval_base(self, coors, diff=0, ori=None,
                    suppress_errors=False, eps=1e-15):
         """
         See :func:`PolySpace.eval_base()`.
         """
-        base = self.eval_ctx.evaluate(coors, diff=diff,
-                                      eps=eps,
-                                      check_errors=not suppress_errors)
+        if diff == 2:
+            base = self._eval_hessian(coors)
+
+        else:
+            base = self.eval_ctx.evaluate(coors, diff=diff,
+                                          eps=eps,
+                                          check_errors=not suppress_errors)
         return base
 
 class LagrangeSimplexPolySpace(LagrangePolySpace):
@@ -425,13 +313,13 @@ class LagrangeSimplexPolySpace(LagrangePolySpace):
 
     def _define_nodes(self):
         # Factorial.
-        fac = lambda n : reduce(lambda a, b : a * (b + 1), xrange(n), 1)
+        fac = lambda n : reduce(lambda a, b : a * (b + 1), range(n), 1)
 
         geometry = self.geometry
         n_v, dim = geometry.n_vertex, geometry.dim
         order = self.order
 
-        n_nod = fac(order + dim) / (fac(order) * fac(dim))
+        n_nod = fac(order + dim) // (fac(order) * fac(dim))
         ## print n_nod, gd
         nodes = nm.zeros((n_nod, n_v), nm.int32)
         nts = nm.zeros((n_nod, 2), nm.int32)
@@ -450,7 +338,10 @@ class LagrangeSimplexPolySpace(LagrangePolySpace):
             nodes[iseq:iseq+n_v,:] = aux
             iseq += n_v
 
-            if dim == 1:
+            if dim == 0:
+                pass
+
+            elif dim == 1:
                 iseq = LagrangeNodes.append_edges(nodes, nts, iseq, 3,
                                                   [[0, 1]], order)
             elif dim == 2:
@@ -486,6 +377,72 @@ class LagrangeSimplexPolySpace(LagrangePolySpace):
 
         return nodes, nts, node_coors
 
+    def _eval_hessian(self, coors):
+        """
+        Evaluate the second derivatives of the basis.
+        """
+        def get_bc(coor):
+            rhs = nm.concatenate((coor, [1]))
+            bc = nm.dot(self.mtx_i, rhs)
+
+            return bc
+
+        def get_val(bc, node, omit=[]):
+            val = nm.ones(1, nm.float64)
+            for i1 in range(bc.shape[0]):
+                if i1 in omit: continue
+
+                for i2 in range(node[i1]):
+                    val *= (self.order * bc[i1] - i2) / (i2 + 1.0)
+
+            return val
+
+        def get_der(bc1, node1, omit=[]):
+            val = nm.zeros(1, nm.float64)
+            for i1 in range(node1):
+                if i1 in omit: continue
+
+                aux = nm.ones(1, nm.float64)
+                for i2 in range(node1):
+                    if (i1 == i2) or (i2 in omit): continue
+                    aux *= (self.order * bc1 - i2) / (i2 + 1.0)
+
+                val += aux * self.order / (i1 + 1.0)
+
+            return val
+
+        n_v = self.mtx_i.shape[0]
+        dim = n_v - 1
+
+        mi = self.mtx_i[:, :dim]
+        bfgg = nm.zeros((coors.shape[0], dim, dim, self.n_nod),
+                        dtype=nm.float64)
+
+        for ic, coor in enumerate(coors):
+            bc = get_bc(coor)
+
+            for ii, node in enumerate(self.nodes):
+                for ig1, bc1 in enumerate(bc): # 1. derivative w.r.t. bc1.
+                    for ig2, bc2 in enumerate(bc): # 2. derivative w.r.t. bc2.
+                        if ig1 == ig2:
+                            val = get_val(bc, node, omit=[ig1])
+
+                            vv = 0.0
+                            for i1 in range(node[ig1]):
+                                aux = get_der(bc2, node[ig2], omit=[i1])
+                                vv += aux * self.order / (i1 + 1.0)
+
+                            val *= vv
+
+                        else:
+                            val = get_val(bc, node, omit=[ig1, ig2])
+                            val *= get_der(bc1, node[ig1])
+                            val *= get_der(bc2, node[ig2])
+
+                        bfgg[ic, :, :, ii] += val * mi[ig1] * mi[ig2][:, None]
+
+        return bfgg
+
 class LagrangeSimplexBPolySpace(LagrangeSimplexPolySpace):
     """Lagrange polynomial space with forced bubble function on a simplex
     domain."""
@@ -541,7 +498,7 @@ class LagrangeTensorProductPolySpace(LagrangePolySpace):
 
         g1d = Struct(n_vertex = 2,
                      dim = 1,
-                     coors = self.bbox[:,0:1])
+                     coors = self.bbox[:,0:1].copy())
         self.ps1d = LagrangeSimplexPolySpace('P_aux', g1d, order,
                                              init_context=False)
 
@@ -578,8 +535,7 @@ class LagrangeTensorProductPolySpace(LagrangePolySpace):
 
             # Vertex nodes.
             nts[0:n_v,0] = 0
-            nts[0:n_v,1] = nm.arange( n_v, dtype = nm.int32 )
-            order * nm.identity( n_v, dtype = nm.int32 )
+            nts[0:n_v,1] = nm.arange(n_v, dtype=nm.int32)
             if dim == 3:
                 for ii in range(n_v):
                     i1, i2, i3 = vertex_map[ii]
@@ -629,7 +585,7 @@ class LagrangeTensorProductPolySpace(LagrangePolySpace):
 
         else:
             c_min, c_max = self.bbox[:,0]
-            
+
             cr = nm.arange(2 * dim)
             node_coors = (nodes[:,cr[::2]] * c_min
                           + nodes[:,cr[1::2]] * c_max) / order
@@ -642,21 +598,21 @@ class LagrangeTensorProductPolySpace(LagrangePolySpace):
         dim = self.geometry.dim
 
         ev = self.ps1d.eval_base
-        
+
         if diff:
             base = nm.ones((coors.shape[0], dim, self.n_nod), dtype=nm.float64)
 
-            for ii in xrange(dim):
+            for ii in range(dim):
                 self.ps1d.nodes = self.nodes[:,2*ii:2*ii+2].copy()
                 self.ps1d.n_nod = self.n_nod
 
-                for iv in xrange(dim):
+                for iv in range(dim):
                     if ii == iv:
                         base[:,iv:iv+1,:] *= ev(coors[:,ii:ii+1].copy(),
                                                 diff=True,
                                                 suppress_errors=suppress_errors,
                                                 eps=eps)
-                
+
                     else:
                         base[:,iv:iv+1,:] *= ev(coors[:,ii:ii+1].copy(),
                                                 diff=False,
@@ -666,10 +622,10 @@ class LagrangeTensorProductPolySpace(LagrangePolySpace):
         else:
             base = nm.ones((coors.shape[0], 1, self.n_nod), dtype=nm.float64)
 
-            for ii in xrange(dim):
+            for ii in range(dim):
                 self.ps1d.nodes = self.nodes[:,2*ii:2*ii+2].copy()
                 self.ps1d.n_nod = self.n_nod
-                
+
                 base *= ev(coors[:,ii:ii+1].copy(),
                            diff=diff,
                            suppress_errors=suppress_errors,
@@ -677,10 +633,188 @@ class LagrangeTensorProductPolySpace(LagrangePolySpace):
 
         return base
 
+    def _eval_hessian(self, coors):
+        """
+        Evaluate the second derivatives of the basis.
+        """
+        evh = self.ps1d.eval_base
+
+        dim = self.geometry.dim
+        bfgg = nm.zeros((coors.shape[0], dim, dim, self.n_nod),
+                        dtype=nm.float64)
+
+        v0s = []
+        v1s = []
+        v2s = []
+        for ii in range(dim):
+            self.ps1d.nodes = self.nodes[:,2*ii:2*ii+2].copy()
+            self.ps1d.n_nod = self.n_nod
+            ev = self.ps1d.create_context(None, 0, 1e-15, 100, 1e-8,
+                                          tdim=1).evaluate
+
+            v0s.append(ev(coors[:, ii:ii+1].copy())[:, 0, :])
+            v1s.append(ev(coors[:, ii:ii+1].copy(), diff=1)[:, 0, :])
+            v2s.append(evh(coors[:, ii:ii+1], diff=2)[:, 0, 0, :])
+
+        for ir in range(dim):
+            vv = v2s[ir] # Destroys v2s!
+            for ik in range(dim):
+                if ik == ir: continue
+                vv *= v0s[ik]
+
+            bfgg[:, ir, ir, :] = vv
+
+            for ic in range(dim):
+                if ic == ir: continue
+                val = v1s[ir] * v1s[ic]
+                for ik in range(dim):
+                    if (ik == ir) or (ik == ic): continue
+                    val *= v0s[ik]
+
+                bfgg[:, ir, ic, :] += val
+
+        return bfgg
+
     def get_mtx_i(self):
         return self.ps1d.mtx_i
 
-class LobattoTensorProductPolySpace(PolySpace):
+class SerendipityTensorProductPolySpace(FEPolySpace):
+    """
+    Serendipity polynomial space using Lagrange functions.
+
+    Notes
+    -----
+    - Orders >= 4 (with bubble functions) are not supported.
+    - Does not use CLagrangeContext, basis functions are hardcoded.
+    - `self.nodes`, `self.node_coors` are not used for basis evaluation and
+       assembling.
+    """
+    name = 'serendipity_tensor_product'
+    supported_orders = {1, 2, 3}
+
+    from sfepy.discrete.fem._serendipity import all_bfs
+
+    def __init__(self, name, geometry, order):
+        import sympy as sm
+
+        if geometry.dim < 2:
+            raise ValueError('serendipity elements need dimension 2 or 3! (%d)'
+                             % geometry.dim)
+        if order not in self.supported_orders:
+            raise ValueError('serendipity elements support only orders %s! (%d)'
+                             % (self.supported_orders, order))
+        PolySpace.__init__(self, name, geometry, order)
+
+        self.nodes, self.nts, self.node_coors = self._define_nodes()
+        self.n_nod = self.nodes.shape[0]
+
+        bfs = self.all_bfs[geometry.dim][order]
+        self.bfs = bfs[0]
+        self.bfgs = bfs[1]
+
+        x, y, z = sm.symbols('x y z')
+        vs = [x, y, z][:geometry.dim]
+
+        self._bfs = [sm.lambdify(vs, bf) for bf in self.bfs]
+        self._bfgs = [[sm.lambdify(vs, bfg) for bfg in bfgs]
+                      for bfgs in self.bfgs]
+
+    def create_context(self, cmesh, eps, check_errors, i_max, newton_eps,
+                       tdim=None):
+        pass
+
+    def _define_nodes(self):
+        geometry = self.geometry
+        order = self.order
+
+        n_v, dim = geometry.n_vertex, geometry.dim
+
+        vertex_map = order * nm.array(vertex_maps[dim], dtype=nm.int32)
+
+        # Only for orders 1, 2, 3!
+        if dim == 2:
+            n_nod = 4 * self.order
+
+        else:
+            n_nod = 8 + 12 * (self.order - 1)
+
+        nodes = nm.zeros((n_nod, 2 * dim), nm.int32)
+        nts = nm.zeros((n_nod, 2), nm.int32)
+
+        if order == 0:
+            nts[0, :] = [3, 0]
+            nodes[0, :] = nm.zeros((n_nod,), nm.int32)
+
+        else:
+            iseq = 0
+
+            # Vertex nodes.
+            nts[0:n_v, 0] = 0
+            nts[0:n_v, 1] = nm.arange(n_v, dtype=nm.int32)
+
+            if dim == 3:
+                for ii in range(n_v):
+                    i1, i2, i3 = vertex_map[ii]
+                    nodes[iseq, :] = [order - i1, i1,
+                                      order - i2, i2,
+                                      order - i3, i3]
+                    iseq += 1
+
+            else: # dim == 2:
+                for ii in range(n_v):
+                    i1, i2 = vertex_map[ii]
+                    nodes[iseq, :] = [order - i1, i1, order - i2, i2]
+                    iseq += 1
+
+            if dim == 2:
+                iseq = LagrangeNodes.append_tp_edges(nodes, nts, iseq, 1,
+                                                     geometry.edges, order)
+
+            elif dim == 3:
+                iseq = LagrangeNodes.append_tp_edges(nodes, nts, iseq, 1,
+                                                     geometry.edges, order)
+
+            else:
+                raise NotImplementedError
+
+        # Coordinates of the nodes.
+        c_min, c_max = self.bbox[:, 0]
+
+        cr = nm.arange(2 * dim)
+        node_coors = (nodes[:, cr[::2]] * c_min
+                      + nodes[:, cr[1::2]] * c_max) / order
+
+        return nodes, nts, nm.ascontiguousarray(node_coors)
+
+    def _eval_base(self, coors, diff=0, ori=None,
+                   suppress_errors=False, eps=1e-15):
+        """
+        See :func:`PolySpace.eval_base()`.
+        """
+        dim = self.geometry.dim
+        if diff:
+            bdim = dim
+
+        else:
+            bdim = 1
+
+        base = nm.empty((coors.shape[0], bdim, self.n_nod), dtype=nm.float64)
+
+        if diff == 0:
+            for ib, bf in enumerate(self._bfs):
+                base[:, 0, ib] = bf(*coors.T)
+
+        elif diff == 1:
+            for ib, bfg in enumerate(self._bfgs):
+                for ig in range(dim):
+                    base[:, ig, ib] = bfg[ig](*coors.T)
+
+        else:
+            raise NotImplementedError
+
+        return base
+
+class LobattoTensorProductPolySpace(FEPolySpace):
     """
     Hierarchical polynomial space using Lobatto functions.
 
@@ -834,7 +968,7 @@ class LobattoTensorProductPolySpace(PolySpace):
         """
         See PolySpace.eval_base().
         """
-        from extmods.lobatto_bases import eval_lobatto_tensor_product as ev
+        from .extmods.lobatto_bases import eval_lobatto_tensor_product as ev
         c_min, c_max = self.bbox[:, 0]
 
         base = ev(coors, self.nodes, c_min, c_max, self.order, diff)
@@ -874,3 +1008,149 @@ class LobattoTensorProductPolySpace(PolySpace):
             base = ebase
 
         return base
+
+class BernsteinSimplexPolySpace(FEPolySpace):
+    """
+    Bernstein polynomial space on simplex domains.
+
+    Notes
+    -----
+    Naive proof-of-concept implementation, does not use recurrent formulas or
+    Duffy transformation to obtain tensor product structure.
+    """
+    name = 'bernstein_simplex'
+
+    def __init__(self, name, geometry, order):
+        PolySpace.__init__(self, name, geometry, order)
+
+        self.nodes, self.nts, self.node_coors = self._define_nodes()
+        self.n_nod = self.nodes.shape[0]
+        self.eval_ctx = None
+
+    def _define_nodes(self):
+        nodes, nts, node_coors = LagrangeSimplexPolySpace._define_nodes(self)
+        return nodes, nts, node_coors
+
+    @staticmethod
+    def _get_barycentric(coors):
+        dim = coors.shape[1]
+
+        bcoors = nm.empty((coors.shape[0], dim + 1))
+        bcoors[:, 0] = 1.0 - coors.sum(axis=1)
+        bcoors[:, 1:] = coors
+        return bcoors
+
+    def _eval_base(self, coors, diff=False, ori=None,
+                   suppress_errors=False, eps=1e-15):
+        """
+        See PolySpace.eval_base().
+        """
+        from scipy.special import factorial
+
+        dim = self.geometry.dim
+        if diff:
+            bdim = dim
+
+            bgrad = nm.zeros((dim + 1, dim), dtype=nm.float64)
+            bgrad[0] = -1
+            bgrad[1:] = nm.eye(dim)
+
+        else:
+            bdim = 1
+
+        base = nm.ones((coors.shape[0], bdim, self.n_nod), dtype=nm.float64)
+
+        bcoors = self._get_barycentric(coors)
+
+        fs = factorial(nm.arange(0, self.order + 1))
+        of = fs[-1]
+        if not diff:
+            for ii, node in enumerate(self.nodes):
+                coef = of / nm.prod(fs[node])
+                val = coef * nm.prod(nm.power(bcoors, node), axis=1)
+                base[:, 0, ii] = val
+
+        else:
+            for ii, node in enumerate(self.nodes):
+                coef = of / nm.prod(fs[node])
+                for ider in range(dim):
+                    dval = 0.0
+                    for ib in range(dim + 1):
+                        ex = node[ib]
+                        val = coef
+                        for im in range(dim + 1):
+                            if ib == im:
+                                val *= (ex *
+                                        nm.power(bcoors[:, im], ex - 1) *
+                                        bgrad[ib, ider])
+
+                            else:
+                                val *= nm.power(bcoors[:, im], node[im])
+
+                        dval += val
+
+                    base[:, ider, ii] = dval
+
+        return base
+
+class BernsteinTensorProductPolySpace(FEPolySpace):
+    """
+    Bernstein polynomial space.
+
+    Each row of the `nodes` attribute defines indices of 1D Bernstein basis
+    functions that need to be multiplied together to evaluate the corresponding
+    shape function. This defines the ordering of basis functions on the
+    reference element.
+    """
+    name = 'bernstein_tensor_product'
+    def __init__(self, name, geometry, order):
+        PolySpace.__init__(self, name, geometry, order)
+
+        self.nodes, self.nts, self.node_coors = self._define_nodes()
+        self.n_nod = self.nodes.shape[0]
+        self.eval_ctx = None
+
+    def _define_nodes(self):
+        nn, nts, node_coors = LagrangeTensorProductPolySpace._define_nodes(self)
+        nodes = nn[:, 1::2]
+
+        return nodes, nts, node_coors
+
+    def _eval_base(self, coors, diff=False, ori=None,
+                   suppress_errors=False, eps=1e-15):
+        """
+        See PolySpace.eval_base().
+        """
+        from sfepy.discrete.iga.extmods.igac import eval_bernstein_basis as ev
+
+        dim = self.geometry.dim
+        if diff:
+            bdim = dim
+
+        else:
+            bdim = 1
+
+        base = nm.ones((coors.shape[0], bdim, self.n_nod), dtype=nm.float64)
+        degree = self.order
+        n_efuns_max = degree + 1
+
+        for iq, qp in enumerate(coors):
+            B = nm.empty((dim, n_efuns_max), dtype=nm.float64)
+            dB_dxi = nm.empty((dim, n_efuns_max), dtype=nm.float64)
+            for ii in range(dim):
+                ev(B[ii, :], dB_dxi[ii, :], qp[ii], degree)
+
+            if not diff:
+                for ii, ni in enumerate(self.nodes.T):
+                    base[iq, 0, :] *= B[ii, ni]
+
+            else:
+                for ii, ni in enumerate(self.nodes.T):
+                    for iv in range(bdim):
+                        if ii == iv:
+                            base[iq, iv, :] *= dB_dxi[ii, ni]
+
+                        else:
+                            base[iq, iv, :] *= B[ii, ni]
+
+        return base
