@@ -1,17 +1,19 @@
+from __future__ import absolute_import
 import os.path as op
 import shutil
 
 import numpy as nm
 
-from sfepy.base.base import output, set_defaults, assert_
+from sfepy.base.base import ordered_iteritems, output, set_defaults, assert_
 from sfepy.base.base import Struct
 from sfepy.homogenization.engine import HomogenizationEngine
-from sfepy.homogenization.homogen_app import get_volume_from_options
 from sfepy.homogenization.homogen_app import HomogenizationApp
 from sfepy.homogenization.coefficients import Coefficients
 from sfepy.homogenization.coefs_base import CoefDummy
 from sfepy.applications import PDESolverApp
 from sfepy.base.plotutils import plt
+import six
+from six.moves import range
 
 def try_set_defaults(obj, attr, defaults, recur=False):
     try:
@@ -22,7 +24,7 @@ def try_set_defaults(obj, attr, defaults, recur=False):
 
     else:
         if recur and isinstance(values, dict):
-            for key, val in values.iteritems():
+            for key, val in six.iteritems(values):
                 set_defaults(val, defaults)
 
         else:
@@ -30,6 +32,19 @@ def try_set_defaults(obj, attr, defaults, recur=False):
 
     return values
 
+def save_raw_bg_logs(filename, logs):
+    """
+    Save raw band gaps `logs` into the `filename` file.
+    """
+    out = {}
+
+    iranges = nm.cumsum([0] + [len(ii) for ii in logs.freqs])
+    out['iranges'] = iranges
+    for key, log in ordered_iteritems(logs.to_dict()):
+        out[key] = nm.concatenate(log, axis=0)
+
+    nm.savez(filename, **out)
+
 def transform_plot_data(datas, plot_transform, conf):
     if plot_transform is not None:
         fun = conf.get_function(plot_transform[0])
@@ -144,9 +159,9 @@ def plot_logs(fig_num, plot_rsc, plot_labels,
         plt.show()
     return fig
 
-def plot_gap(ax, ii, f0, f1, kind, kind_desc, gmin, gmax, plot_range, plot_rsc):
+def plot_gap(ax, ranges, kind, kind_desc, plot_range, plot_rsc):
     """
-    Plot a single band gap as a rectangle.
+    Plot single band gap frequency ranges as rectangles.
     """
     def draw_rect(ax, x, y, rsc):
         ax.fill(nm.asarray(x)[[0,1,1,0]],
@@ -159,42 +174,29 @@ def plot_gap(ax, ii, f0, f1, kind, kind_desc, gmin, gmax, plot_range, plot_rsc):
     propagation = plot_rsc['propagation']
 
     if kind == 'p':
-        draw_rect(ax, (f0, f1), plot_range, propagation)
-        info = [(f0, f1)]
+        draw_rect(ax, ranges[0], plot_range, propagation)
     elif kind == 'w':
-        draw_rect(ax, (f0, f1), plot_range, weak)
-        info = [(f0, f1)]
+        draw_rect(ax, ranges[0], plot_range, weak)
     elif kind == 'wp':
-        draw_rect(ax, (f0, gmin[1]), plot_range, weak)
-        draw_rect(ax, (gmin[1], f1), plot_range, propagation)
-        info = [(f0, gmin[1]), (gmin[1], f1)]
+        draw_rect(ax, ranges[0], plot_range, weak)
+        draw_rect(ax, ranges[1], plot_range, propagation)
     elif kind == 's':
-        draw_rect(ax, (f0, f1), plot_range, strong)
-        info = [(f0, f1)]
+        draw_rect(ax, ranges[0], plot_range, strong)
     elif kind == 'sw':
-        draw_rect(ax, (f0, gmax[1]), plot_range, strong)
-        draw_rect(ax, (gmax[1], f1), plot_range, weak)
-        info = [(f0, gmax[1]), (gmax[1], f1)]
+        draw_rect(ax, ranges[0], plot_range, strong)
+        draw_rect(ax, ranges[1], plot_range, weak)
     elif kind == 'swp':
-        draw_rect(ax, (f0, gmax[1]), plot_range, strong)
-        draw_rect(ax, (gmax[1], gmin[1]), plot_range, weak)
-        draw_rect(ax, (gmin[1], f1), plot_range, propagation)
-        info = [(f0, gmax[1]), (gmax[1], gmin[1]), (gmin[1], f1)]
+        draw_rect(ax, ranges[0], plot_range, strong)
+        draw_rect(ax, ranges[1], plot_range, weak)
+        draw_rect(ax, ranges[2], plot_range, propagation)
     elif kind == 'is':
-        draw_rect(ax, (gmin[1], gmax[1]), plot_range, strong)
-        info = [(gmin[1], gmax[1])]
+        draw_rect(ax, ranges[0], plot_range, strong)
     elif kind == 'iw':
-        draw_rect(ax, (gmin[1], gmax[1]), plot_range, weak)
-        info = [(gmin[1], gmax[1])]
+        draw_rect(ax, ranges[0], plot_range, weak)
     else:
-        output('impossible band gap combination:')
-        output(gmin, gmax)
-        raise ValueError
+        raise ValueError('unknown band gap kind! (%s)' % kind)
 
-    output(ii, gmin[0], gmax[0], '%.8f' % f0, '%.8f' % f1)
-    output(' -> %s\n    %s' %(kind_desc, info))
-
-def plot_gaps(fig_num, plot_rsc, gaps, kinds, freq_range,
+def plot_gaps(fig_num, plot_rsc, gaps, kinds, gap_ranges, freq_range,
               plot_range, show=False, clear=False, new_axes=False):
     """
     Plot band gaps as rectangles.
@@ -209,20 +211,24 @@ def plot_gaps(fig_num, plot_rsc, gaps, kinds, freq_range,
     else:
         ax = fig.gca()
 
-    for ii in xrange(len(freq_range) - 1):
+    for ii in range(len(freq_range) - 1):
         f0, f1 = freq_range[[ii, ii+1]]
         gap = gaps[ii]
+        ranges = gap_ranges[ii]
 
         if isinstance(gap, list):
             for ig, (gmin, gmax) in enumerate(gap):
                 kind, kind_desc = kinds[ii][ig]
-                plot_gap(ax, ii, f0, f1, kind, kind_desc, gmin, gmax,
-                         plot_range, plot_rsc)
+                plot_gap(ax, ranges[ig], kind, kind_desc, plot_range, plot_rsc)
+
+                output(ii, gmin[0], gmax[0], '%.8f' % f0, '%.8f' % f1)
+                output(' -> %s\n    %s' %(kind_desc, ranges[ig]))
         else:
             gmin, gmax = gap
             kind, kind_desc = kinds[ii]
-            plot_gap(ax, ii, f0, f1, kind, kind_desc, gmin, gmax,
-                     plot_range, plot_rsc)
+            plot_gap(ax, ranges, kind, kind_desc, plot_range, plot_rsc)
+            output(ii, gmin[0], gmax[0], '%.8f' % f0, '%.8f' % f1)
+            output(' -> %s\n    %s' %(kind_desc, ranges))
 
     if new_axes:
         ax.set_xlim([freq_range[0], freq_range[-1]])
@@ -305,9 +311,9 @@ class AcousticBandGapsApp(HomogenizationApp):
             'weak_gap' : {'linewidth' : 0, 'facecolor' : (0.6, 0.8, 0.6)},
             'propagation' : {'linewidth' : 0, 'facecolor' : (1, 1, 1)},
             'params' : {'axes.labelsize': 'x-large',
-                        'text.fontsize': 'large',
+                        'font.size': 14,
                         'legend.fontsize': 'large',
-                        'legend.loc': 0,
+                        'legend.loc': 'best',
                         'xtick.labelsize': 'large',
                         'ytick.labelsize': 'large',
                         'text.usetex': True},
@@ -351,9 +357,10 @@ class AcousticBandGapsApp(HomogenizationApp):
 
         self.setup_options()
 
-        output_dir = self.problem.output_dir
-        shutil.copyfile(conf._filename,
-                        op.join(output_dir, op.basename(conf._filename)))
+        if conf._filename:
+            output_dir = self.problem.output_dir
+            shutil.copyfile(conf._filename,
+                            op.join(output_dir, op.basename(conf._filename)))
 
     def setup_options(self):
         HomogenizationApp.setup_options(self)
@@ -384,7 +391,7 @@ class AcousticBandGapsApp(HomogenizationApp):
         elif options.analyze_dispersion or options.phase_velocity:
 
             # Insert incident wave direction to coefficients that need it.
-            for key, val in coef_info.iteritems():
+            for key, val in six.iteritems(coef_info):
                 coef_opts = val.get('options', None)
                 if coef_opts is None: continue
 
@@ -415,16 +422,23 @@ class AcousticBandGapsApp(HomogenizationApp):
 
         he_options = Struct(coefs=coefs_name, requirements=opts.requirements,
                             compute_only=keys,
-                            post_process_hook=self.post_process_hook)
-        volume = get_volume_from_options(opts, self.problem)
+                            post_process_hook=self.post_process_hook,
+                            multiprocessing=False)
+
+        volumes = {}
+        if hasattr(opts, 'volumes') and (opts.volumes is not None):
+            volumes.update(opts.volumes)
+        elif hasattr(opts, 'volume') and (opts.volume is not None):
+            volumes['total'] = opts.volume
+        else:
+            volumes['total'] = 1.0
 
         he = HomogenizationEngine(self.problem, options,
                                   app_options=he_options,
-                                  volume=volume)
+                                  volumes=volumes)
         coefs = he()
 
         coefs = Coefficients(**coefs.to_dict())
-        coefs.volume = volume
 
         coefs_filename = op.join(opts.output_dir, opts.coefs_filename)
         coefs.to_file_txt(coefs_filename + '.txt',
@@ -441,6 +455,11 @@ class AcousticBandGapsApp(HomogenizationApp):
                 filename = op.join(self.problem.output_dir, log_save_name)
                 bg.save_log(filename, opts.float_format, bg)
 
+            raw_log_save_name = bg.get('raw_log_save_name', None)
+            if raw_log_save_name is not None:
+                filename = op.join(self.problem.output_dir, raw_log_save_name)
+                save_raw_bg_logs(filename, bg.logs)
+
         if options.plot:
             if options.detect_band_gaps:
                 self.plot_band_gaps(coefs)
@@ -475,7 +494,7 @@ class AcousticBandGapsApp(HomogenizationApp):
             plot_range, teigs = transform_plot_data(bg.logs.eigs,
                                                     opts.plot_transform,
                                                     self.conf)
-            fig = plot_gaps(ii, plot_rsc, bg.gaps, bg.kinds,
+            fig = plot_gaps(ii, plot_rsc, bg.gaps, bg.kinds, bg.gap_ranges,
                             bg.freq_range_margins, plot_range,
                             clear=True)
             fig = plot_logs(ii, plot_rsc, plot_labels, bg.logs.freqs, teigs,
@@ -484,6 +503,7 @@ class AcousticBandGapsApp(HomogenizationApp):
                             plot_range,
                             show_legend=plot_opts['legend'],
                             new_axes=True)
+            plt.tight_layout()
 
             if opts.fig_name is not None:
                 fig_name = _get_fig_name(self.problem.output_dir, opts.fig_name,
@@ -517,7 +537,7 @@ class AcousticBandGapsApp(HomogenizationApp):
 
             bg = coefs.get(key)
 
-            fig = plot_gaps(1, plot_rsc, bg.gaps, bg.kinds,
+            fig = plot_gaps(1, plot_rsc, bg.gaps, bg.kinds, bg.gap_ranges,
                             bg.freq_range_margins, plot_range,
                             clear=True)
             fig = plot_logs(1, plot_rsc, plot_labels, bg.logs.freqs, pas,
@@ -526,6 +546,7 @@ class AcousticBandGapsApp(HomogenizationApp):
                             plot_range,
                             show_legend=plot_opts['legend'],
                             new_axes=True)
+            plt.tight_layout()
 
             fig_name = opts.fig_name_angle
             if fig_name is not None:
@@ -540,7 +561,7 @@ class AcousticBandGapsApp(HomogenizationApp):
 
             plot_labels =  opts.plot_labels_wave
 
-            fig = plot_gaps(2, plot_rsc, bg.gaps, bg.kinds,
+            fig = plot_gaps(2, plot_rsc, bg.gaps, bg.kinds, bg.gap_ranges,
                             bg.freq_range_margins, plot_range,
                             clear=True)
             fig = plot_logs(2, plot_rsc, plot_labels, bg.logs.freqs, teigs,
@@ -549,6 +570,7 @@ class AcousticBandGapsApp(HomogenizationApp):
                             plot_range,
                             show_legend=plot_opts['legend'],
                             new_axes=True)
+            plt.tight_layout()
 
             fig_name = opts.fig_name_wave
             if fig_name is not None:
