@@ -1,5 +1,8 @@
+from __future__ import print_function
+from __future__ import absolute_import
 import numpy as nm
 import sys
+from six.moves import range
 sys.path.append('.')
 
 from sfepy.base.base import output, assert_
@@ -106,7 +109,8 @@ def gen_block_mesh(dims, shape, centre, mat_id=0, name='block',
     ngrid = nm.mgrid[[slice(ii) for ii in shape]]
     ngrid.shape = (dim, n_nod)
 
-    coors = x0 + ngrid.T * dd
+    if coors is None:
+        coors = x0 + ngrid.T * dd
     output('...done', verbose=verbose)
 
     n_el = nm.prod(shape - 1)
@@ -410,8 +414,8 @@ def tiled_mesh1d(conn, coors, ngrps, idim, n_rep, bb, eps=1e-6, ndmap=False):
     s2 = nm.nonzero(coors[:,idim] > (bb[1] - eps))[0]
 
     if s1.shape != s2.shape:
-        raise ValueError, 'incompatible shapes: %s == %s'\
-              % (s1.shape, s2.shape)
+        raise ValueError('incompatible shapes: %s == %s'\
+              % (s1.shape, s2.shape))
 
     (nnod0, dim) = coors.shape
     nnod = nnod0 * n_rep - s1.shape[0] * (n_rep - 1)
@@ -683,7 +687,7 @@ def gen_mesh_from_geom(geo, a=None, verbose=False, refine=False):
     params += " %s" % (polyfilename)
 
     cmd = "%s %s" % (meshgen_call[geo.dim][0], params)
-    if verbose: print "Generating mesh using", cmd
+    if verbose: print("Generating mesh using", cmd)
 
     p=pexpect.run(cmd, timeout=None)
     bname, ext = op.splitext(polyfilename)
