@@ -1,18 +1,29 @@
 """
 Functions for a high-level PETSc-based parallelization.
 """
-import time
+from __future__ import absolute_import
+import os
 
 import numpy as nm
+from six.moves import range
 
-import sys, petsc4py
-argv = [arg for arg in sys.argv if arg not in ['-h', '--help']]
-petsc4py.init(argv)
+def init_petsc_args():
+    try:
+        import sys, petsc4py
+
+    except ImportError:
+        return
+
+    argv = [arg for arg in sys.argv if arg not in ['-h', '--help']]
+    petsc4py.init(argv)
+
+init_petsc_args()
 
 from petsc4py import PETSc
 from mpi4py import MPI
 
 from sfepy.base.base import assert_, output, ordered_iteritems, Struct
+from sfepy.base.timing import Timer
 from sfepy.discrete.common.region import Region
 from sfepy.discrete.fem.fe_surface import FESurface
 
@@ -22,7 +33,7 @@ def partition_mesh(mesh, n_parts, use_metis=True, verbose=False):
     available.
     """
     output('partitioning mesh into %d subdomains...' % n_parts, verbose=verbose)
-    tt = time.clock()
+    timer = Timer(start=True)
 
     if use_metis:
         try:
@@ -43,7 +54,7 @@ def partition_mesh(mesh, n_parts, use_metis=True, verbose=False):
 
     else:
         ii = nm.arange(n_parts)
-        n_cell_parts = mesh.n_el / n_parts + ((mesh.n_el % n_parts) > ii)
+        n_cell_parts = mesh.n_el // n_parts + ((mesh.n_el % n_parts) > ii)
         output('cell counts:', n_cell_parts, verbose=verbose)
         assert_(sum(n_cell_parts) == mesh.n_el)
         assert_(nm.all(n_cell_parts > 0))
@@ -51,7 +62,7 @@ def partition_mesh(mesh, n_parts, use_metis=True, verbose=False):
         offs = nm.cumsum(nm.r_[0, n_cell_parts])
         cell_tasks = nm.digitize(nm.arange(offs[-1]), offs) - 1
 
-    output('...done in', time.clock() - tt, verbose=verbose)
+    output('...done in', timer.stop(), verbose=verbose)
 
     return cell_tasks
 
@@ -99,7 +110,8 @@ def get_inter_facets(domain, cell_tasks):
     return inter_facets
 
 def create_task_dof_maps(field, cell_tasks, inter_facets, is_overlap=True,
-                         use_expand_dofs=False):
+                         use_expand_dofs=False, save_inter_regions=False,
+                         output_dir=None):
     """
     For each task list its inner and interface DOFs of the given field and
     create PETSc numbering that is consecutive in each subdomain.
@@ -116,6 +128,11 @@ def create_task_dof_maps(field, cell_tasks, inter_facets, is_overlap=True,
     cells may be added even if not needed - filter them out by using the PETSc
     DOFs range.
 
+    When debugging domain partitioning problems, it is advisable to set
+    `save_inter_regions` to True to save the task interfaces as meshes as well
+    as vertex-based markers - to be used only with moderate problems and small
+    numbers of tasks.
+
     [1] J. Sistek and F. Cirak. Parallel iterative solution of the
     incompressible Navier-Stokes equations with application to rotating
     wings. Submitted for publication, 2015
@@ -166,6 +183,20 @@ def create_task_dof_maps(field, cell_tasks, inter_facets, is_overlap=True,
                                         parent=cregion.name)
             region.update_shape()
 
+            if save_inter_regions:
+                output_dir = output_dir if output_dir is not None else '.'
+                filename = os.path.join(output_dir, '%s.mesh' % name)
+
+                aux = domain.mesh.from_region(region, domain.mesh,
+                                              is_surface=True)
+                aux.write(filename, io='auto')
+
+                mask = nm.zeros((domain.mesh.n_nod, 1), dtype=nm.float64)
+                mask[region.vertices] = 1
+                out = {name : Struct(name=name, mode='vertex', data=mask)}
+                filename = os.path.join(output_dir, '%s.h5' % name)
+                domain.mesh.write(filename, out=out, io='auto')
+
             inter_dofs.append(_get_dofs_region(field, region))
 
             sd = FESurface('surface_data_%s' % region.name, region,
@@ -265,7 +296,7 @@ def verify_task_dof_maps(dof_maps, id_map, field, use_expand_dofs=False,
 
     Returns the vector with a task number for each DOF.
     """
-    tt = time.clock()
+    timer = Timer(start=True)
     if verbose:
         output('verifying...')
         output('total number of DOFs:', field.n_nod)
@@ -321,7 +352,7 @@ def verify_task_dof_maps(dof_maps, id_map, field, use_expand_dofs=False,
     assert_(n_dof == len(id_map))
     assert_(n_dof == len(nm.unique(id_map)))
 
-    output('...done in', time.clock() - tt, verbose=verbose)
+    output('...done in', timer.stop(), verbose=verbose)
 
     return vec
 
@@ -345,7 +376,7 @@ def distribute_field_dofs(field, gfd, use_expand_dofs=False,
         id_map = gfd.id_map
 
         # Send subdomain data to other tasks.
-        for it in xrange(1, size):
+        for it in range(1, size):
             # Send owned and overlap cells.
             cells = nm.union1d(gfd.cell_parts[it], gfd.overlap_cells[it])
             mpi.send(len(cells), it)
@@ -397,17 +428,19 @@ def distribute_field_dofs(field, gfd, use_expand_dofs=False,
         dof_maps = id_map = None
 
     if verbose:
+        output('field %s:' % field.name)
         output('n_cell:', n_cell)
         output('cells:', cells)
         output('owned petsc DOF range:', petsc_dofs_range,
                petsc_dofs_range[1] - petsc_dofs_range[0])
         aux = nm.unique(petsc_dofs_conn)
-        output('local petsc DOFs (owned + shared):', aux, len(aux))
+        output('%d local petsc DOFs (owned + shared):' % len(aux), aux)
 
     return cells, petsc_dofs_range, petsc_dofs_conn, dof_maps, id_map
 
 def distribute_fields_dofs(fields, cell_tasks, is_overlap=True,
-                           use_expand_dofs=False, comm=None, verbose=False):
+                           use_expand_dofs=False, save_inter_regions=False,
+                           output_dir=None, comm=None, verbose=False):
     """
     Distribute the owned cells and DOFs of the given field to all tasks.
 
@@ -428,13 +461,15 @@ def distribute_fields_dofs(fields, cell_tasks, is_overlap=True,
         for field in fields:
             aux = create_task_dof_maps(field, cell_tasks, inter_facets,
                                        is_overlap=is_overlap,
-                                       use_expand_dofs=use_expand_dofs)
+                                       use_expand_dofs=use_expand_dofs,
+                                       save_inter_regions=save_inter_regions,
+                                       output_dir=output_dir)
             cell_parts = aux[2]
             n_cell_parts = [len(ii) for ii in cell_parts]
             output('numbers of cells in tasks (without overlaps):',
                    n_cell_parts, verbose=verbose)
             assert_(sum(n_cell_parts) == field.domain.mesh.n_el)
-            assert_(nm.all(n_cell_parts > 0))
+            assert_(nm.all(nm.array(n_cell_parts) > 0))
 
             gfd = Struct(name='global field %s distribution' % field.name,
                          dof_maps=aux[0], id_map=aux[1],
@@ -446,7 +481,7 @@ def distribute_fields_dofs(fields, cell_tasks, is_overlap=True,
         if len(fields) > 1:
             # Renumber id_maps for field inter-leaving.
             offset = 0
-            for ir in xrange(size):
+            for ir in range(size):
                 for ii, gfd in enumerate(gfds):
                     dof_map = gfd.dof_maps[ir]
                     n_owned = dof_map[3]
@@ -460,7 +495,7 @@ def distribute_fields_dofs(fields, cell_tasks, is_overlap=True,
 
         else:
             gfd = gfds[0]
-            gfd.coffsets[:] = [gfd.dof_maps[ir][4] for ir in xrange(size)]
+            gfd.coffsets[:] = [gfd.dof_maps[ir][4] for ir in range(size)]
 
     else:
         gfds = [None] * len(fields)
@@ -539,14 +574,15 @@ def setup_composite_dofs(lfds, fields, local_variables, verbose=False):
         lfd.petsc_dofs = get_local_ordering(variable.field,
                                             lfd.petsc_dofs_conn,
                                             use_expand_dofs=True)
-        output('petsc dofs:', lfd.petsc_dofs, verbose=verbose)
+        output('%d petsc dofs:' % len(lfd.petsc_dofs), lfd.petsc_dofs,
+               verbose=verbose)
 
     sizes, drange = get_composite_sizes(lfds)
     output('composite sizes:', sizes, verbose=verbose)
     output('composite drange:', drange, verbose=verbose)
 
     pdofs = nm.concatenate([ii.petsc_dofs for ii in lfds])
-    output('composite pdofs:', pdofs, verbose=verbose)
+    output('%d composite pdofs:' % len(pdofs), pdofs, verbose=verbose)
 
     return sizes, drange, pdofs
 
@@ -562,7 +598,7 @@ def expand_dofs(dofs, n_components):
         sh = None
 
     edofs = nm.empty(n_components * dofs.shape[0], nm.int32)
-    for idof in xrange(n_components):
+    for idof in range(n_components):
         aux = n_components * dofs + idof
         edofs[idof::n_components] = aux
 
@@ -571,19 +607,28 @@ def expand_dofs(dofs, n_components):
 
     return edofs
 
+def call_in_rank_order(fun, comm=None):
+    """
+    Call a function `fun` task by task in the task rank order.
+    """
+    if comm is None:
+        comm = PETSc.COMM_WORLD
+
+    for rank in range(comm.size):
+        if rank == comm.rank:
+            fun(rank, comm)
+        comm.barrier()
+
 def view_petsc_local(data, name='data', viewer=None, comm=None):
     """
     View local PETSc `data` called `name`. The data object has to have
     `.view()` method.
     """
-    if comm is None:
-        comm = PETSc.COMM_WORLD
+    def _view(rank, comm):
+        output('contents of local %s on rank %d:' % (name, rank))
+        data.view(viewer=viewer)
 
-    for rank in xrange(comm.size):
-        if rank == comm.rank:
-            output('contents of local %s on rank %d:' % (name, rank))
-            data.view(viewer=viewer)
-        comm.barrier()
+    call_in_rank_order(_view, comm=comm)
 
 def create_local_petsc_vector(pdofs):
     """
@@ -648,7 +693,8 @@ def create_prealloc_data(mtx, pdofs, drange, verbose=False):
     """
     owned_dofs = nm.where((pdofs >= drange[0]) & (pdofs < drange[1]))[0]
     owned_dofs = owned_dofs.astype(nm.int32)
-    output('owned local DOFs:', owned_dofs, verbose=verbose)
+    output('%d owned local DOFs:' % len(owned_dofs), owned_dofs,
+           verbose=verbose)
 
     ii = nm.argsort(pdofs[owned_dofs])
     aux = mtx[owned_dofs[ii]]
@@ -706,19 +752,6 @@ def create_petsc_system(mtx, sizes, pdofs, drange, is_overlap=True,
 
     return pmtx, psol, prhs
 
-def apply_ebc_to_matrix(mtx, ebc_rows):
-    """
-    Apply to matrix rows: zeros to non-diagonal entries, one to the diagonal.
-    """
-    data, prows, cols = mtx.data, mtx.indptr, mtx.indices
-    for ir in ebc_rows:
-        for ic in xrange(prows[ir], prows[ir + 1]):
-            if (cols[ic] == ir):
-                data[ic] = 1.0
-
-            else:
-                data[ic] = 0.0
-
 def assemble_rhs_to_petsc(prhs, rhs, pdofs, drange, is_overlap=True,
                           comm=None, verbose=False):
     """
@@ -727,80 +760,69 @@ def assemble_rhs_to_petsc(prhs, rhs, pdofs, drange, is_overlap=True,
     if comm is None:
         comm = PETSc.COMM_WORLD
 
-    lgmap = PETSc.LGMap().create(pdofs, comm=comm)
-
+    timer = Timer()
     if is_overlap:
-        prhs.setLGMap(lgmap)
         output('setting rhs values...', verbose=verbose)
-        tt = time.clock()
-        for ir, rdof in enumerate(pdofs):
-            if (rdof < drange[0]) or (rdof >= drange[1]): continue
-            prhs.setValueLocal(ir, rhs[ir],
-                               PETSc.InsertMode.INSERT_VALUES)
-        output('...done in', time.clock() - tt, verbose=verbose)
+        timer.start()
+        rdofs = nm.where((pdofs < drange[0]) | (pdofs >= drange[1]), -1, pdofs)
+        prhs.setOption(prhs.Option.IGNORE_NEGATIVE_INDICES, True)
+        prhs.setValues(rdofs, rhs, PETSc.InsertMode.INSERT_VALUES)
+        output('...done in', timer.stop(), verbose=verbose)
 
         output('assembling rhs...', verbose=verbose)
-        tt = time.clock()
+        timer.start()
         prhs.assemble()
-        output('...done in', time.clock() - tt, verbose=verbose)
+        output('...done in', timer.stop(), verbose=verbose)
 
     else:
-        prhs.setLGMap(lgmap)
         output('setting rhs values...', verbose=verbose)
-        tt = time.clock()
-        prhs.setValuesLocal(nm.arange(len(rhs), dtype=nm.int32), rhs,
-                            PETSc.InsertMode.ADD_VALUES)
-        output('...done in', time.clock() - tt, verbose=verbose)
+        timer.start()
+        prhs.setValues(pdofs, rhs, PETSc.InsertMode.ADD_VALUES)
+        output('...done in', timer.stop(), verbose=verbose)
 
         output('assembling rhs...', verbose=verbose)
-        tt = time.clock()
+        timer.start()
         prhs.assemble()
-        output('...done in', time.clock() - tt, verbose=verbose)
+        output('...done in', timer.stop(), verbose=verbose)
 
 def assemble_mtx_to_petsc(pmtx, mtx, pdofs, drange, is_overlap=True,
                           comm=None, verbose=False):
     """
     Assemble a local CSR matrix to a global PETSc matrix.
-
-    WIP
-    ---
-    Try Mat.setValuesCSR() - no lgmap - filtering vectorized?
     """
     if comm is None:
         comm = PETSc.COMM_WORLD
 
+    timer = Timer()
+
     lgmap = PETSc.LGMap().create(pdofs, comm=comm)
-
+    pmtx.setLGMap(lgmap, lgmap)
     if is_overlap:
-        pmtx.setLGMap(lgmap, lgmap)
-
-        data, prows, cols = mtx.data, mtx.indptr, mtx.indices
-
         output('setting matrix values...', verbose=verbose)
-        tt = time.clock()
-        for ir, rdof in enumerate(pdofs):
-            if (rdof < drange[0]) or (rdof >= drange[1]): continue
-
-            for ic in xrange(prows[ir], prows[ir + 1]):
-                # output(ir, rdof, cols[ic])
-                pmtx.setValueLocal(ir, cols[ic], data[ic],
-                                   PETSc.InsertMode.INSERT_VALUES)
-        output('...done in', time.clock() - tt, verbose=verbose)
+        timer.start()
+        mask = (pdofs < drange[0]) | (pdofs >= drange[1])
+        nnz_per_row = nm.diff(mtx.indptr)
+        mtx2 = mtx.copy()
+        mtx2.data[nm.repeat(mask, nnz_per_row)] = 0
+        mtx2.eliminate_zeros()
+        pmtx.setValuesLocalCSR(mtx2.indptr, mtx2.indices, mtx2.data,
+                               PETSc.InsertMode.INSERT_VALUES)
+        output('...done in', timer.stop(), verbose=verbose)
 
         output('assembling matrix...', verbose=verbose)
-        tt = time.clock()
+        timer.start()
         pmtx.assemble()
-        output('...done in', time.clock() - tt, verbose=verbose)
+        output('...done in', timer.stop(), verbose=verbose)
+
 
     else:
-        pmtx.setLGMap(lgmap, lgmap)
         output('setting matrix values...', verbose=verbose)
-        tt = time.clock()
+        timer.start()
         pmtx.setValuesLocalCSR(mtx.indptr, mtx.indices, mtx.data,
                                PETSc.InsertMode.ADD_VALUES)
-        output('...done in', time.clock() - tt, verbose=verbose)
+        output('...done in', timer.stop(), verbose=verbose)
 
         output('assembling matrix...', verbose=verbose)
-        tt = time.clock()
+        timer.start()
         pmtx.assemble()
-        output('...done in', time.clock() - tt, verbose=verbose)
+        output('...done in', timer.stop(), verbose=verbose)
