@@ -1,7 +1,10 @@
+from __future__ import print_function
+from __future__ import absolute_import
 import os
 import shutil, tempfile
 
 import numpy as nm
+from six.moves import range
 
 try:
     from enthought.traits.api \
@@ -13,6 +16,7 @@ try:
     from enthought.mayavi.tools.mlab_scene_model import MlabSceneModel
     from enthought.mayavi.core.ui.mayavi_scene import MayaviScene
     from enthought.mayavi.core.dataset_manager import DatasetManager
+    from enthought.pyface.gui import GUI
 
 except ImportError:
     from traits.api \
@@ -24,6 +28,7 @@ except ImportError:
     from mayavi.tools.mlab_scene_model import MlabSceneModel
     from mayavi.core.ui.mayavi_scene import MayaviScene
     from mayavi.core.dataset_manager import DatasetManager
+    from pyface.gui import GUI
 
 from sfepy.base.base import (insert_as_static_method, output, assert_,
                              get_arguments, get_default, Struct, basestr)
@@ -342,12 +347,12 @@ class Viewer(Struct):
 
     def encode_animation(self, filename, format, ffmpeg_options=None):
         if ffmpeg_options is None:
-            ffmpeg_options = '-r 10 -sameq'
+            ffmpeg_options = '-framerate 10'
 
         base, _, ext = self.get_animation_info(filename)
         anim_name = '.'.join((base, format))
-        cmd = 'ffmpeg %s -i %s %s' \
-              % (ffmpeg_options, '.'.join((base, '%*', ext[1:])), anim_name)
+        cmd = 'ffmpeg %s -pattern_type glob -i "%s" -q:v 0 %s' \
+              % (ffmpeg_options, '.'.join((base, '*', ext[1:])), anim_name)
         output('creating animation "%s"...' % anim_name)
         output('using command:')
         output(cmd)
@@ -403,6 +408,8 @@ class Viewer(Struct):
 
         else:
             is_subdomains = False
+            if 'mat_id' not in filter_names:
+                file_source.setup_mat_id()
 
         self.source = source = self.file_source()
         data_ranges = get_data_ranges(source, return_only=True)
@@ -472,7 +479,7 @@ class Viewer(Struct):
             for ij, iric in enumerate(cycle((n_row, n_col))):
                 ir, ic = iric
                 if ij < len(repeat):
-                    for ik in xrange(repeat[ij]):
+                    for ik in range(repeat[ij]):
                         yield ii, ir, ic
                         ii += 1
 
@@ -1193,7 +1200,10 @@ def make_animation(filename, view, roll, anim_format, options,
 
 class ClosingHandler(Handler):
     def object_button_quit_changed(self, info):
+        if not info.initialized: return
         info.ui.dispose()
+        gui = GUI()
+        gui.stop_event_loop()
 
 class ViewerGUI(HasTraits):
 
@@ -1232,10 +1242,10 @@ class ViewerGUI(HasTraits):
         self.scene.camera.print_traits()
         view = mlab.view()
         roll = mlab.roll()
-        print 'view:', view
-        print 'roll:', roll
-        print 'as args: --view=%.2e,%.2e,%.2e,%.2e,%.2e,%.2e --roll=%.2e' \
-              % (view[:3] + tuple(view[3]) + (roll,))
+        print('view:', view)
+        print('roll:', roll)
+        print('as args: --view=%.2e,%.2e,%.2e,%.2e,%.2e,%.2e --roll=%.2e' \
+              % (view[:3] + tuple(view[3]) + (roll,)))
 
     def _button_make_animation_steps_fired(self):
         view = mlab.view()
