@@ -13,9 +13,7 @@ def configuration(parent_package='',top_path=None):
         'homogenization',
         'linalg',
         'mechanics',
-        'optimize',
         'parallel',
-        'physics',
         'postprocess',
         'solvers',
         'terms'
