@@ -1,14 +1,17 @@
 """
 Nonlinear solvers.
 """
-import time
+from __future__ import absolute_import
 
 import numpy as nm
 import numpy.linalg as nla
 
 from sfepy.base.base import output, get_default, debug, Struct
 from sfepy.base.log import Log, get_logging_conf
-from sfepy.solvers.solvers import SolverMeta, NonlinearSolver
+from sfepy.base.timing import Timer
+from sfepy.solvers.solvers import NonlinearSolver
+import six
+from six.moves import range
 
 def check_tangent_matrix(conf, vec_x0, fun, fun_grad):
     """
@@ -46,14 +49,14 @@ def check_tangent_matrix(conf, vec_x0, fun, fun_grad):
 
     vec_r = fun(vec_x) # Restore.
 
-    tt = time.clock()
+    timer = Timer(start=True)
     output(mtx_a, '.. analytical')
     output(mtx_d, '.. difference')
     import sfepy.base.plotutils as plu
     plu.plot_matrix_diff(mtx_d, mtx_a, delta, ['difference', 'analytical'],
                          conf.check)
 
-    return time.clock() - tt
+    return timer.stop()
 
 def conv_test(conf, it, err, err0):
     """
@@ -84,9 +87,9 @@ def conv_test(conf, it, err, err0):
 
     output('nls: iter: %d, residual: %e (rel: %e)' % (it, err, err_r))
 
-    conv_a = err < conf.eps_a
+    conv_a = err <= conf.eps_a
     if it > 0:
-        conv_r = err_r < conf.eps_r
+        conv_r = err_r <= conf.eps_r
 
         if conv_a and conv_r:
             status = 0
@@ -105,13 +108,12 @@ def conv_test(conf, it, err, err0):
 
 class Newton(NonlinearSolver):
     r"""
-    Solves a nonlinear system :math:`f(x) = 0` using the Newton method with
-    backtracking line-search, starting with an initial guess :math:`x^0`.
+    Solves a nonlinear system :math:`f(x) = 0` using the Newton method.
+
+    The solver uses a backtracking line-search on divergence.
     """
     name = 'nls.newton'
 
-    __metaclass__ = SolverMeta
-
     _parameters = [
         ('i_max', 'int', 1, False,
          'The maximum number of iterations.'),
@@ -226,7 +228,9 @@ class Newton(NonlinearSolver):
         eps_r = get_default(ls_eps_r, 1.0)
         lin_red = conf.eps_a * conf.lin_red
 
-        time_stats = {}
+        timer = Timer()
+        time_stats_keys = ['residual', 'matrix', 'solve']
+        time_stats = {key : 0.0 for key in time_stats_keys}
 
         vec_x = vec_x0.copy()
         vec_x_last = vec_x0.copy()
@@ -238,14 +242,16 @@ class Newton(NonlinearSolver):
         err = err0 = -1.0
         err_last = -1.0
         it = 0
+        ls_status = {}
+        ls_n_iter = 0
         while 1:
             if iter_hook is not None:
-                iter_hook(self, vec_x, it, err, err0)
+                iter_hook(self.context, self, vec_x, it, err, err0)
 
             ls = 1.0
-            vec_dx0 = vec_dx;
+            vec_dx0 = vec_dx
             while 1:
-                tt = time.clock()
+                timer.start()
 
                 try:
                     vec_r = fun(vec_x)
@@ -261,7 +267,7 @@ class Newton(NonlinearSolver):
                 else:
                     ok = True
 
-                time_stats['rezidual'] = time.clock() - tt
+                time_stats['residual'] = timer.stop()
                 if ok:
                     try:
                         err = nla.norm(vec_r)
@@ -286,7 +292,7 @@ class Newton(NonlinearSolver):
                         break
 
                     red = conf.ls_red_warp;
-                    output('rezidual computation failed for iter %d'
+                    output('residual computation failed for iter %d'
                            ' (new ls: %e)!' % (it, red * ls))
 
                 if ls < conf.ls_min:
@@ -313,19 +319,19 @@ class Newton(NonlinearSolver):
                 condition = 2
                 break
 
-            tt = time.clock()
+            timer.start()
             if not conf.is_linear:
                 mtx_a = fun_grad(vec_x)
 
             else:
                 mtx_a = fun_grad('linear')
 
-            time_stats['matrix'] = time.clock() - tt
+            time_stats['matrix'] = timer.stop()
 
             if conf.check:
-                tt = time.clock()
+                timer.start()
                 wt = check_tangent_matrix(conf, vec_x, fun, fun_grad)
-                time_stats['check'] = time.clock() - tt - wt
+                time_stats['check'] = timer.stop() - wt
 
             if conf.lin_precision is not None:
                 if ls_eps_a is not None:
@@ -339,16 +345,18 @@ class Newton(NonlinearSolver):
             if conf.verbose:
                 output('solving linear system...')
 
-            tt = time.clock()
+            timer.start()
             vec_dx = lin_solver(vec_r, x0=vec_x,
-                                eps_a=eps_a, eps_r=eps_r, mtx=mtx_a)
-            time_stats['solve'] = time.clock() - tt
+                                eps_a=eps_a, eps_r=eps_r, mtx=mtx_a,
+                                status=ls_status)
+            ls_n_iter += ls_status['n_iter']
+            time_stats['solve'] = timer.stop()
 
             if conf.verbose:
                 output('...done')
 
-            for kv in time_stats.iteritems():
-                output('%10s: %7.2f [s]' % kv)
+            for key in time_stats_keys:
+                output('%10s: %7.2f [s]' % (key, time_stats[key]))
 
             vec_e = mtx_a * vec_dx - vec_r
             lerr = nla.norm(vec_e)
@@ -365,6 +373,7 @@ class Newton(NonlinearSolver):
             status['err0'] = err0
             status['err'] = err
             status['n_iter'] = it
+            status['ls_n_iter'] = ls_n_iter if ls_n_iter >= 0 else -1
             status['condition'] = condition
 
         if conf.log.plot is not None:
@@ -379,8 +388,6 @@ class ScipyBroyden(NonlinearSolver):
     """
     name = 'nls.scipy_broyden_like'
 
-    __metaclass__ = SolverMeta
-
     _parameters = [
         ('method', 'str', 'anderson', False,
          'The name of the solver in ``scipy.optimize``.'),
@@ -420,7 +427,7 @@ class ScipyBroyden(NonlinearSolver):
         fun = get_default(fun, self.fun)
         status = get_default(status, self.status)
 
-        tt = time.clock()
+        timer = Timer(start=True)
 
         kwargs = {'iter' : conf.i_max,
                   'alpha' : conf.alpha,
@@ -440,7 +447,7 @@ class ScipyBroyden(NonlinearSolver):
         vec_x = nm.asarray(vec_x)
 
         if status is not None:
-            status['time_stats'] = time.clock() - tt
+            status['time_stats'] = timer.stop()
 
         return vec_x
 
@@ -459,8 +466,6 @@ class PETScNonlinearSolver(NonlinearSolver):
     """
     name = 'nls.petsc'
 
-    __metaclass__ = SolverMeta
-
     _parameters = [
         ('method', 'str', 'newtonls', False,
          'The SNES type.'),
@@ -490,8 +495,21 @@ class PETScNonlinearSolver(NonlinearSolver):
 
         from petsc4py import PETSc as petsc
 
+        converged_reasons = {}
+        for key, val in six.iteritems(petsc.SNES.ConvergedReason.__dict__):
+            if isinstance(val, int):
+                converged_reasons[val] = key
+
+        ksp_converged_reasons = {}
+        for key, val in six.iteritems(petsc.KSP.ConvergedReason.__dict__):
+            if isinstance(val, int):
+                ksp_converged_reasons[val] = key
+
         NonlinearSolver.__init__(self, conf, petsc=petsc,
-                                 pmtx=pmtx, prhs=prhs, comm=comm, **kwargs)
+                                 pmtx=pmtx, prhs=prhs, comm=comm,
+                                 converged_reasons=converged_reasons,
+                                 ksp_converged_reasons=ksp_converged_reasons,
+                                 **kwargs)
 
     def __call__(self, vec_x0, conf=None, fun=None, fun_grad=None,
                  lin_solver=None, iter_hook=None, status=None,
@@ -506,7 +524,7 @@ class PETScNonlinearSolver(NonlinearSolver):
         prhs = get_default(prhs, self.prhs)
         comm = get_default(comm, self.comm)
 
-        tt = time.clock()
+        timer = Timer(start=True)
 
         if isinstance(vec_x0, self.petsc.Vec):
             psol = vec_x0
@@ -533,10 +551,54 @@ class PETScNonlinearSolver(NonlinearSolver):
         snes.setMaxFunctionEvaluations(conf.if_max)
         snes.setFromOptions()
 
+        fun(snes, psol, prhs)
+        err0 = prhs.norm()
+
         snes.solve(prhs.duplicate(), psol)
 
         if status is not None:
-            status['time_stats'] = time.clock() - tt
+            status['time_stats'] = timer.stop()
+
+        if snes.reason in self.converged_reasons:
+            reason = 'snes: %s' % self.converged_reasons[snes.reason]
+
+        else:
+            reason = 'ksp: %s' % self.ksp_converged_reasons[snes.reason]
+
+        output('%s(%s): %d iterations in the last step'
+               % (ksp.getType(), ksp.getPC().getType(),
+                  ksp.getIterationNumber()),
+               verbose=conf.verbose)
+
+        output('%s convergence: %s (%s, %d iterations, %d function evaluations)'
+               % (snes.getType(), snes.reason, reason,
+                  snes.getIterationNumber(), snes.getFunctionEvaluations()),
+               verbose=conf.verbose)
+
+        converged = snes.reason >= 0
+
+        if not converged:
+            # PETSc does not update the solution if KSP have not converged.
+            dpsol = snes.getSolutionUpdate()
+            psol -= dpsol
+
+            fun(snes, psol, prhs)
+            err = prhs.norm()
+
+        else:
+            try:
+                err = snes.getFunctionNorm()
+
+            except AttributeError:
+                fun(snes, psol, prhs)
+                err = prhs.norm()
+
+        if status is not None:
+            status['err0'] = err0
+            status['err'] = err
+            status['n_iter'] = snes.getIterationNumber()
+            status['ls_n_iter'] = snes.getLinearSolveIterations()
+            status['condition'] = 0 if converged else -1
 
         if isinstance(vec_x0, self.petsc.Vec):
             sol = psol
