@@ -1,3 +1,4 @@
+from __future__ import absolute_import
 import re
 from copy import copy
 
@@ -9,6 +10,9 @@ from sfepy.base.compat import in1d
 
 # Used for imports in term files.
 from sfepy.terms.extmods import terms
+import six
+from six.moves import range
+from functools import reduce
 
 _match_args = re.compile('^([^\(\}]*)\((.*)\)$').match
 _match_virtual = re.compile('^virtual$').match
@@ -145,6 +149,7 @@ def create_arg_parser():
     from pyparsing import Literal, Word, delimitedList, Group, \
          StringStart, StringEnd, Optional, nums, alphas, alphanums
 
+    ident = Word(alphas, alphanums + "_")
     inumber = Word("+-" + nums, nums)
 
     history = Optional(Literal('[').suppress() + inumber
@@ -156,7 +161,10 @@ def create_arg_parser():
     derivative = Group(Literal('d') + variable\
                        + Literal('/').suppress() + Literal('dt'))
 
-    trace = Group(Literal('tr') + Literal('(').suppress() + variable \
+    trace = Group(Literal('tr')
+                  + Literal('(').suppress()
+                  + Optional(ident + Literal(',').suppress(), default=None)
+                  + variable
                   + Literal(')').suppress())
 
     generalized_var = derivative | trace | variable
@@ -165,17 +173,18 @@ def create_arg_parser():
 
     return args
 
+
 class ConnInfo(Struct):
 
     def get_region(self, can_trace=True):
         if self.is_trace and can_trace:
-            return self.region.get_mirror_region()
+            return self.region.get_mirror_region(self.trace_region)
         else:
             return self.region
 
     def get_region_name(self, can_trace=True):
         if self.is_trace and can_trace:
-            reg = self.region.get_mirror_region()
+            reg = self.region.get_mirror_region(self.trace_region)
         else:
             reg = self.region
 
@@ -349,12 +358,18 @@ class Term(Struct):
 
         return obj
 
+    @staticmethod
+    def tile_mat(mat, nel):
+        if mat.shape[0] == 1 and nel > 1:
+            return nm.tile(mat, (nel,) + (mat.ndim - 1) * (1,))
+        else:
+            return mat
+
     def __init__(self, name, arg_str, integral, region, **kwargs):
         self.name = name
         self.arg_str = arg_str
         self.region = region
         self._kwargs = kwargs
-        self._integration = self.integration
         self.sign = 1.0
 
         self.set_integral(integral)
@@ -400,6 +415,11 @@ class Term(Struct):
         out = -1.0 * self
         return out
 
+    def get_str(self):
+        return '{:+} * {}.{}.{}({})'.format(
+            self.sign, self.name, self.integral.order,
+            self.region.name, self.arg_str)
+
     def set_integral(self, integral):
         """
         Set the term integral.
@@ -429,24 +449,27 @@ class Term(Struct):
         self.arg_steps = {}
         self.arg_derivatives = {}
         self.arg_traces = {}
+        self.arg_trace_regions = {}
 
         parser = create_arg_parser()
         self.arg_desc = parser.parseString(self.arg_str)
-
         for arg in self.arg_desc:
-            trace = False
             derivative = None
+            trace = False
+            trace_region = None
 
             if isinstance(arg[1], int):
                 name, step = arg
 
             else:
                 kind = arg[0]
-                name, step = arg[1]
                 if kind == 'd':
+                    name, step = arg[1]
                     derivative = arg[2]
                 elif kind == 'tr':
                     trace = True
+                    trace_region = arg[1]
+                    name, step = arg[2]
 
             match = _match_material_root(name)
             if match:
@@ -456,6 +479,7 @@ class Term(Struct):
             self.arg_steps[name] = step
             self.arg_derivatives[name] = derivative
             self.arg_traces[name] = trace
+            self.arg_trace_regions[name] = trace_region
 
     def setup_args(self, **kwargs):
         self._kwargs = kwargs
@@ -480,6 +504,8 @@ class Term(Struct):
         if user is None:
             user = {}
 
+        user.setdefault('ts', Struct())
+
         kwargs = {}
         for arg_name in self.arg_names:
             if isinstance(arg_name, basestr):
@@ -576,6 +602,10 @@ class Term(Struct):
                     self.names.parameter.append(name)
 
             elif arg_kind.endswith('material'):
+                # This should be better checked already in create_arg_parser().
+                if not isinstance(name, tuple):
+                    raise ValueError('wrong material argument %s of term %s!'
+                                     % (name, self.get_str()))
                 names = self.names.material
 
             else:
@@ -620,7 +650,19 @@ class Term(Struct):
             if field is None:
                 continue
 
-            if not nm.all(in1d(self.region.vertices,
+            region = self.region
+            if self.arg_traces[name]:
+                mreg_name = self.arg_trace_regions[name]
+                if mreg_name is None:
+                    mreg_name = region.setup_mirror_region(mreg_name,
+                                                           ret_name=True)
+                    self.arg_trace_regions[name] = mreg_name
+                else:
+                    region.setup_mirror_region(mreg_name)
+
+                region = region.get_mirror_region(mreg_name)
+
+            if not nm.all(in1d(region.vertices,
                                field.region.vertices)):
                 msg = ('%s: incompatible regions: (self, field %s)'
                        + '(%s in %s)') %\
@@ -662,7 +704,14 @@ class Term(Struct):
     def get_conn_key(self):
         """The key to be used in DOF connectivity information."""
         key = (self.name,) + tuple(self.arg_names)
-        key += (self.integral_name, self.region.name)
+        arg_traces = [k for k, v in self.arg_traces.items() if v]
+        if len(arg_traces) > 0:
+            atr = arg_traces[-1]
+            trace = True, self.arg_trace_regions[atr], atr
+        else:
+            trace = False, None, None
+
+        key += (self.integral_name, self.region.name) + trace
 
         return key
 
@@ -693,10 +742,16 @@ class Term(Struct):
 
         region = self.get_region()
         if region is not None:
-            is_any_trace = reduce(lambda x, y: x or y,
-                                  self.arg_traces.values())
-            if is_any_trace:
-                region.setup_mirror_region()
+            arg_traces = [k for k, v in self.arg_traces.items() if v]
+            if len(arg_traces) > 0:
+                aname = arg_traces[-1]
+                mreg_name = self.arg_trace_regions[aname]
+                if mreg_name is None:
+                    mreg_name = region.setup_mirror_region(mreg_name,
+                                                           ret_name=True)
+                    self.arg_trace_regions[aname] = mreg_name
+                else:
+                    region.setup_mirror_region(mreg_name)
 
         vals = []
         aux_pvars = []
@@ -708,6 +763,7 @@ class Term(Struct):
 
             field = svar.get_field()
             is_trace = self.arg_traces[svar.name]
+            trace_region = self.arg_trace_regions[svar.name]
 
             if svar.name in tgs:
                 ps_tg = tgs[svar.name]
@@ -720,6 +776,7 @@ class Term(Struct):
                            has_virtual=True,
                            has_state=True,
                            is_trace=is_trace,
+                           trace_region=trace_region,
                            dc_type=dc_type,
                            v_tg=v_tg,
                            ps_tg=ps_tg,
@@ -731,6 +788,7 @@ class Term(Struct):
         for pvar in pvars:
             field = pvar.get_field()
             is_trace = self.arg_traces[pvar.name]
+            trace_region = self.arg_trace_regions[pvar.name]
 
             if pvar.name in tgs:
                 ps_tg = tgs[pvar.name]
@@ -743,6 +801,7 @@ class Term(Struct):
                            has_virtual=vvar is not None,
                            has_state=False,
                            is_trace=is_trace,
+                           trace_region=trace_region,
                            dc_type=dc_type,
                            v_tg=v_tg,
                            ps_tg=ps_tg,
@@ -758,6 +817,7 @@ class Term(Struct):
                            has_virtual=True,
                            has_state=False,
                            is_trace=False,
+                           trace_region=None,
                            dc_type=dc_type,
                            v_tg=v_tg,
                            ps_tg=v_tg,
@@ -863,6 +923,15 @@ class Term(Struct):
         self.has_geometry = True
 
         self.geometry_types = {}
+        if self.integration == 'by_region':
+            reg = self.region
+            if reg.kind == 'cell':
+                integration = 'volume'
+            elif reg.kind == 'face' or reg.kind == 'facet':
+                integration = getattr(self, 'surface_integration', 'surface')
+
+            self.integration = integration
+
         if isinstance(self.integration, basestr):
             for var in self.get_variables():
                 self.geometry_types[var.name] = self.integration
@@ -872,11 +941,11 @@ class Term(Struct):
                 self.integration = self._integration[self.mode]
 
             if self.integration is not None:
-                for arg_type, gtype in self.integration.iteritems():
+                for arg_type, gtype in six.iteritems(self.integration):
                     var = self.get_args(arg_types=[arg_type])[0]
                     self.geometry_types[var.name] = gtype
 
-        gtypes = list(set(self.geometry_types.itervalues()))
+        gtypes = list(set(self.geometry_types.values()))
 
         if 'surface_extra' in gtypes:
             self.dof_conn_type = 'volume'
@@ -914,6 +983,9 @@ class Term(Struct):
             self.dt = ts.dt
             self.is_quasistatic = ts.is_quasistatic
 
+        if 'ts' in self._kwargs:
+            self._kwargs['ts'].update(ts)
+
     def advance(self, ts):
         """
         Advance to the next time step. Implemented in subclasses.
@@ -928,18 +1000,6 @@ class Term(Struct):
         return variable(step=self.arg_steps[name],
                         derivative=self.arg_derivatives[name])
 
-    def get_approximation(self, variable, get_saved=False):
-        """
-        Return approximation corresponding to `variable`. Also return
-        the corresponding geometry (actual or saved, according to
-        `get_saved`).
-        """
-        geo, _, key = self.get_mapping(variable, get_saved=get_saved,
-                                       return_key=True)
-        ap = variable.get_approximation()
-
-        return ap, geo
-
     def get_variables(self, as_list=True):
 
         if as_list:
@@ -1021,7 +1081,8 @@ class Term(Struct):
         is_trace = self.arg_traces[variable.name]
 
         if is_trace:
-            region = self.region.get_mirror_region()
+            mreg_name = self.arg_trace_regions[variable.name]
+            region = self.region.get_mirror_region(mreg_name)
 
         else:
             region = self.region
@@ -1046,7 +1107,8 @@ class Term(Struct):
         is_trace = self.arg_traces[variable.name]
 
         if is_trace:
-            region = self.region.get_mirror_region()
+            mreg_name = self.arg_trace_regions[variable.name]
+            region = self.region.get_mirror_region(mreg_name)
 
         else:
             region = self.region
@@ -1075,7 +1137,8 @@ class Term(Struct):
                                  region=self.region, integral=self.integral,
                                  integration=integration,
                                  step=step, time_derivative=time_derivative,
-                                 is_trace=self.arg_traces[name], bf=bf)
+                                 is_trace=self.arg_traces[name], bf=bf,
+                                 trace_region=self.arg_trace_regions[name])
         return data
 
     def check_shapes(self, *args, **kwargs):
@@ -1093,12 +1156,18 @@ class Term(Struct):
                 if sh == 'D':
                     return dim
 
+                elif sh == 'D2':
+                    return dim**2
+
                 elif sh == 'S':
                     return sym
 
                 elif sh == 'N': # General number.
                     return nm.inf
 
+                elif sh == 'str':
+                    return 'str'
+
                 else:
                     return int(sh)
 
@@ -1115,13 +1184,20 @@ class Term(Struct):
 
         arg_kinds = get_arg_kinds(self.ats)
 
-        arg_shapes_list = self.arg_shapes
+        if hasattr(self, 'arg_shapes_dict') and \
+                isinstance(self.arg_shapes_dict, dict):
+            integration = self.integration.split('_')[0]
+            arg_shapes_list = self.arg_shapes_dict[integration]
+        else:
+            arg_shapes_list = self.arg_shapes
+
         if not isinstance(arg_shapes_list, list):
             arg_shapes_list = [arg_shapes_list]
 
         # Loop allowed shapes until a match is found, else error.
         allowed_shapes = []
         prev_shapes = {}
+        actual_shapes = {}
         for _arg_shapes in arg_shapes_list:
             # Unset shapes are taken from the previous iteration.
             arg_shapes = copy(prev_shapes)
@@ -1137,6 +1213,7 @@ class Term(Struct):
                     continue
 
                 arg = args[ii]
+                key = '%s:%s' % (self.ats[ii], self.arg_names[ii])
 
                 if self.mode is not None:
                     extended_ats = self.ats[ii] + ('/%s' % self.mode)
@@ -1152,6 +1229,7 @@ class Term(Struct):
 
                 if arg_kind.endswith('variable'):
                     n_el, n_qp, _dim, n_en, n_c = self.get_data_shape(arg)
+                    actual_shapes[key] = (n_c,)
                     shape = _parse_scalar_shape(sh[0] if isinstance(sh, tuple)
                                                 else sh)
                     if nm.isinf(shape):
@@ -1170,14 +1248,19 @@ class Term(Struct):
 
                     prefix = ''
                     if isinstance(sh, basestr):
-                        aux = sh.split(':')
+                        aux = tuple(ii.strip() for ii in sh.split(':'))
                         if len(aux) == 2:
                             prefix, sh = aux
 
+                    if sh == 'str':
+                        n_ok += isinstance(arg, basestr)
+                        continue
+
                     shape = _parse_tuple_shape(sh)
                     ls = len(shape)
 
                     aarg = nm.array(arg, ndmin=1)
+                    actual_shapes[key] = aarg.shape
 
                     # Substiture general dimension 'N' with actual value.
                     iinfs = nm.where(nm.isinf(shape))[0]
@@ -1190,11 +1273,12 @@ class Term(Struct):
                     if (ls > 1) or (shape[0] > 1):
                         # Array.
                         n_ok += shape == aarg.shape[-ls:]
+                        actual_shapes[key] = aarg.shape[-ls:]
 
                     elif (ls == 1) and (shape[0] == 1):
-                        # Scalar constant.
+                        # Scalar constant or callable as term argument
                         from numbers import Number
-                        n_ok += isinstance(arg, Number)
+                        n_ok += isinstance(arg, Number) or callable(arg)
 
                 else:
                     n_ok += 1
@@ -1203,10 +1287,11 @@ class Term(Struct):
                 break
 
         else:
-            term_str = '%s.%d.%s(%s)' % (self.name, self.integral.order,
-                                         self.region.name, self.arg_str)
+            term_str = self.get_str()
             output('allowed argument shapes for term "%s":' % term_str)
             output(allowed_shapes)
+            output('actual argument shapes:')
+            output(actual_shapes)
             raise ValueError('wrong arguments shapes for "%s" term! (see above)'
                              % term_str)
 
@@ -1228,9 +1313,9 @@ class Term(Struct):
         try:
             fargs = self.get_fargs(*args, **kwargs)
 
-        except RuntimeError:
+        except (RuntimeError, ValueError):
             terms.errclear()
-            raise ValueError
+            raise
 
         return fargs
 
@@ -1238,9 +1323,9 @@ class Term(Struct):
         try:
             status = self.function(out, *fargs)
 
-        except RuntimeError:
+        except (RuntimeError, ValueError):
             terms.errclear()
-            raise ValueError
+            raise
 
         if status:
             terms.errclear()
@@ -1334,35 +1419,44 @@ class Term(Struct):
 
             emode = 'eval' if mode == 'el_eval' else mode
             _args = tuple(args) + (emode, term_mode, diff_var)
-            fargs = self.call_get_fargs(_args, kwargs)
 
             shape, dtype = self.get_eval_shape(*_args, **kwargs)
 
-            if dtype == nm.float64:
-                val, status = self.eval_real(shape, fargs, mode, term_mode,
-                                             **kwargs)
-
-            elif dtype == nm.complex128:
-                val, status = self.eval_complex(shape, fargs, mode, term_mode,
-                                                **kwargs)
+            if shape[0] == 0:
+                val = nm.zeros(shape, dtype=dtype)
+                status = 0
 
             else:
-                raise ValueError('unsupported term dtype! (%s)' % dtype)
+                fargs = self.call_get_fargs(_args, kwargs)
+
+                if dtype == nm.float64:
+                    val, status = self.eval_real(shape, fargs, mode,
+                                                 term_mode,
+                                                 **kwargs)
+
+                elif dtype == nm.complex128:
+                    val, status = self.eval_complex(shape, fargs, mode,
+                                                    term_mode,
+                                                    **kwargs)
+
+                else:
+                    raise ValueError('unsupported term dtype! (%s)' % dtype)
 
             val *= self.sign
             out = (val,)
 
         elif mode == 'weak':
             varr = self.get_virtual_variable()
+            if varr is None:
+                raise ValueError('no virtual variable in weak mode! (in "%s")'
+                                 % self.get_str())
+
             if diff_var is not None:
                 varc = self.get_variables(as_list=False)[diff_var]
 
             args = self.get_args(**kwargs)
             self.check_shapes(*args)
 
-            _args = tuple(args) + (mode, term_mode, diff_var)
-            fargs = self.call_get_fargs(_args, kwargs)
-
             n_elr, n_qpr, dim, n_enr, n_cr = self.get_data_shape(varr)
             n_row = n_cr * n_enr
 
@@ -1375,27 +1469,41 @@ class Term(Struct):
 
                 shape = (n_elr, 1, n_row, n_col)
 
-            if varr.dtype == nm.float64:
-                vals, status = self.eval_real(shape, fargs, mode, term_mode,
-                                              diff_var, **kwargs)
-
-            elif varr.dtype == nm.complex128:
-                vals, status = self.eval_complex(shape, fargs, mode, term_mode,
-                                                 diff_var, **kwargs)
+            if shape[0] == 0:
+                vals = nm.zeros(shape, dtype=varr.dtype)
+                status = 0
 
             else:
-                raise ValueError('unsupported term dtype! (%s)'
-                                 % varr.dtype)
+                _args = tuple(args) + (mode, term_mode, diff_var)
+                fargs = self.call_get_fargs(_args, kwargs)
+
+                if varr.dtype == nm.float64:
+                    vals, status = self.eval_real(shape, fargs, mode,
+                                                  term_mode,
+                                                  diff_var, **kwargs)
+
+                elif varr.dtype == nm.complex128:
+                    vals, status = self.eval_complex(shape, fargs, mode,
+                                                     term_mode,
+                                                     diff_var, **kwargs)
+
+                else:
+                    raise ValueError('unsupported term dtype! (%s)'
+                                     % varr.dtype)
+
+            if not isinstance(vals, tuple):
+                vals *= self.sign
+                iels = self.get_assembling_cells(vals.shape)
+
+            else:
+                vals = (self.sign * vals[0],) + vals[1:]
+                iels = None
 
-            vals *= self.sign
-            iels = self.get_assembling_cells(vals.shape)
             out = (vals, iels)
 
         if goptions['check_term_finiteness']:
             assert_(nm.isfinite(out[0]).all(),
-                    msg='%+.2e * %s.%d.%s(%s) term values not finite!'
-                    % (self.sign, self.name, self.integral.order,
-                       self.region.name, self.arg_str))
+                    msg='"%s" term values not finite!' % self.get_str())
 
         if ret_status:
             out = out + (status,)
@@ -1406,11 +1514,26 @@ class Term(Struct):
         return out
 
     def assemble_to(self, asm_obj, val, iels, mode='vector', diff_var=None):
+        """
+        Assemble the results of term evaluation.
+
+        For standard terms, assemble the values in `val` corresponding to
+        elements/cells `iels` into a vector or a CSR sparse matrix `asm_obj`,
+        depending on `mode`.
+
+        For terms with a dynamic connectivity (e.g. contact terms), in
+        `'matrix'` mode, return the extra COO sparse matrix instead. The extra
+        matrix has to be added to the global matrix by the caller. By default,
+        this is done in :func:`Equations.evaluate()
+        <sfepy.discrete.equations.Equations.evaluate()>`.
+        """
         import sfepy.discrete.common.extmods.assemble as asm
 
         vvar = self.get_virtual_variable()
         dc_type = self.get_dof_conn_type()
 
+        extra = None
+
         if mode == 'vector':
             if asm_obj.dtype == nm.float64:
                 assemble = asm.assemble_vector
@@ -1422,10 +1545,23 @@ class Term(Struct):
                     if not(val[ii].dtype == nm.complex128):
                         val[ii] = nm.complex128(val[ii])
 
-            dc = vvar.get_dof_conn(dc_type)
-            assert_(val.shape[2] == dc.shape[1])
+            if not isinstance(val, tuple):
+                dc = vvar.get_dof_conn(dc_type)
+                assert_(val.shape[2] == dc.shape[1])
 
-            assemble(asm_obj, val, iels, 1.0, dc)
+                assemble(asm_obj, val, iels, 1.0, dc)
+
+            else:
+                vals, rows, var = val
+                if var.eq_map is not None:
+                    eq = var.eq_map.eq
+
+                    rows = eq[rows]
+                    active = (rows >= 0)
+                    vals, rows = vals[active], rows[active]
+
+                # Assumes no repeated indices in rows!
+                asm_obj[rows] += vals
 
         elif mode == 'matrix':
             if asm_obj.dtype == nm.float64:
@@ -1442,12 +1578,6 @@ class Term(Struct):
                 and (val.dtype == nm.float64)):
                 val = val.astype(nm.complex128)
 
-            rdc = vvar.get_dof_conn(dc_type)
-
-            is_trace = self.arg_traces[svar.name]
-            cdc = svar.get_dof_conn(dc_type, is_trace=is_trace)
-            assert_(val.shape[2:] == (rdc.shape[1], cdc.shape[1]))
-
             sign = 1.0
             if self.arg_derivatives[svar.name]:
                 if not self.is_quasistatic or (self.step > 0):
@@ -1456,7 +1586,31 @@ class Term(Struct):
                 else:
                     sign = 0.0
 
-            assemble(tmd[0], tmd[1], tmd[2], val, iels, sign, rdc, cdc)
+            if not isinstance(val, tuple):
+                rdc = vvar.get_dof_conn(dc_type)
+
+                is_trace = self.arg_traces[svar.name]
+                trace_region = self.arg_trace_regions[svar.name]
+                cdc = svar.get_dof_conn(dc_type, is_trace, trace_region)
+                assert_(val.shape[2:] == (rdc.shape[1], cdc.shape[1]))
+
+                assemble(tmd[0], tmd[1], tmd[2], val, iels, sign, rdc, cdc)
+
+            else:
+                from scipy.sparse import coo_matrix
+
+                vals, rows, cols, rvar, cvar = val
+                if rvar.eq_map is not None:
+                    req, ceq = rvar.eq_map.eq, cvar.eq_map.eq
+
+                    rows, cols = req[rows], ceq[cols]
+                    active = (rows >= 0) & (cols >= 0)
+                    vals, rows, cols = vals[active], rows[active], cols[active]
+
+                extra = coo_matrix((sign * vals, (rows, cols)),
+                                   shape=asm_obj.shape)
 
         else:
             raise ValueError('unknown assembling mode! (%s)' % mode)
+
+        return extra
