@@ -134,11 +134,11 @@ class SUPGCAdjStabilizationTerm(Term):
 
     def get_fargs(self, mat, virtual, state, parameter,
                   mode=None, term_mode=None, diff_var=None, **kwargs):
-        ap, vg = self.get_approximation(state)
+        vg, _ = self.get_mapping(state)
 
         val_u = self.get(parameter, 'val')
         grad_u = self.get(parameter, 'grad').transpose((0, 1, 3, 2)).copy()
-        conn = ap.get_connectivity(self.region, self.integration)
+        conn = state.field.get_connectivity(self.region, self.integration)
 
         fmode = diff_var is not None
 
@@ -169,10 +169,10 @@ class SUPGPAdj1StabilizationTerm(Term):
 
     def get_fargs(self, mat, virtual, state, parameter,
                   mode=None, term_mode=None, diff_var=None, **kwargs):
-        ap_w, vg_w = self.get_approximation(state)
+        vg_w, _ = self.get_mapping(state)
 
         grad_p = self.get(parameter, 'grad')
-        conn_w = ap_w.get_connectivity(self.region, self.integration)
+        conn_w = state.field.get_connectivity(self.region, self.integration)
 
         fmode = diff_var is not None
 
@@ -204,35 +204,35 @@ class SUPGPAdj2StabilizationTerm(Term):
 
     def get_fargs(self, mat, virtual, parameter, state,
                   mode=None, term_mode=None, diff_var=None, **kwargs):
-        ap_r, vg_r = self.get_approximation(state)
+        vg_r, _ = self.get_mapping(state)
         vg_u, _ = self.get_mapping(parameter)
 
         grad_u = self.get(parameter, 'grad').transpose((0, 1, 3, 2)).copy()
-        conn_r = ap_r.get_connectivity(self.region, self.integration)
+        conn_r = state.field.get_connectivity(self.region, self.integration)
 
         fmode = diff_var is not None
 
         return grad_u, state(), mat, vg_u, vg_r, conn_r, fmode
 
-class SDDotVolumeTerm(Term):
+class SDDotTerm(Term):
     r"""
     Sensitivity (shape derivative) of dot product of scalars or vectors.
 
     :Definition:
 
     .. math::
-        \int_{\Omega_D} p q (\nabla \cdot \ul{\Vcal}) \mbox{ , }
-        \int_{\Omega_D} (\ul{u} \cdot \ul{w}) (\nabla \cdot \ul{\Vcal})
+        \int_{\Omega} p q (\nabla \cdot \ul{\Vcal}) \mbox{ , }
+        \int_{\Omega} (\ul{u} \cdot \ul{w}) (\nabla \cdot \ul{\Vcal})
 
     :Arguments:
         - parameter_1 : :math:`p` or :math:`\ul{u}`
         - parameter_2 : :math:`q` or :math:`\ul{w}`
-        - parameter_mesh_velocity : :math:`\ul{\Vcal}`
+        - parameter_mv : :math:`\ul{\Vcal}`
     """
-    name = 'd_sd_volume_dot'
-    arg_types = ('parameter_1', 'parameter_2', 'parameter_mesh_velocity')
+    name = 'ev_sd_dot'
+    arg_types = ('parameter_1', 'parameter_2', 'parameter_mv')
     arg_shapes = [{'parameter_1' : 'D', 'parameter_2' : 'D',
-                   'parameter_mesh_velocity' : 'D'},
+                   'parameter_mv' : 'D'},
                   {'parameter_1' : 1, 'parameter_2' : 1}]
 
     function = staticmethod(terms.d_sd_volume_dot)
@@ -262,18 +262,18 @@ class SDDivTerm(Term):
     :Definition:
 
     .. math::
-        \int_{\Omega_D} p [ (\nabla \cdot \ul{w}) (\nabla \cdot \ul{\Vcal})
+        \int_{\Omega} p [ (\nabla \cdot \ul{w}) (\nabla \cdot \ul{\Vcal})
         - \pdiff{\Vcal_k}{x_i} \pdiff{w_i}{x_k} ]
 
     :Arguments:
         - parameter_u : :math:`\ul{u}`
         - parameter_p : :math:`p`
-        - parameter_mesh_velocity : :math:`\ul{\Vcal}`
+        - parameter_mv : :math:`\ul{\Vcal}`
     """
-    name = 'd_sd_div'
-    arg_types = ('parameter_u', 'parameter_p', 'parameter_mesh_velocity')
+    name = 'ev_sd_div'
+    arg_types = ('parameter_u', 'parameter_p', 'parameter_mv')
     arg_shapes = {'parameter_u' : 'D', 'parameter_p' : 1,
-                  'parameter_mesh_velocity' : 'D'}
+                  'parameter_mv' : 'D'}
 
     function = staticmethod(terms.d_sd_div)
 
@@ -305,28 +305,28 @@ class SDDivGradTerm(Term):
     :Definition:
 
     .. math::
-        w \nu \int_{\Omega_D} [ \pdiff{u_i}{x_k} \pdiff{w_i}{x_k}
+        \nu \int_{\Omega} [ \pdiff{u_i}{x_k} \pdiff{w_i}{x_k}
         (\nabla \cdot \ul{\Vcal})
         - \pdiff{\Vcal_j}{x_k} \pdiff{u_i}{x_j} \pdiff{w_i}{x_k}
         - \pdiff{u_i}{x_k} \pdiff{\Vcal_l}{x_k} \pdiff{w_i}{x_k} ]
 
     :Arguments:
-        - material_1  : :math:`w` (weight)
-        - material_2  : :math:`\nu` (viscosity)
+        - material  : :math:`\nu` (viscosity, optional)
         - parameter_u : :math:`\ul{u}`
         - parameter_w : :math:`\ul{w}`
-        - parameter_mesh_velocity : :math:`\ul{\Vcal}`
+        - parameter_mv : :math:`\ul{\Vcal}`
     """
-    name = 'd_sd_div_grad'
-    arg_types = ('material_1', 'material_2', 'parameter_u', 'parameter_w',
-                 'parameter_mesh_velocity')
-    arg_shapes = {'material_1' : '1, 1', 'material_2' : '1, 1',
-                  'parameter_u' : 'D', 'parameter_w' : 'D',
-                  'parameter_mesh_velocity' : 'D'}
+    name = 'ev_sd_div_grad'
+    arg_types = ('opt_material', 'parameter_u', 'parameter_w',
+                 'parameter_mv')
+    arg_shapes = [{'opt_material' : '1, 1',
+                   'parameter_u' : 'D', 'parameter_w' : 'D',
+                   'parameter_mv' : 'D'},
+                  {'opt_material' : None}]
 
     function = staticmethod(terms.d_sd_div_grad)
 
-    def get_fargs(self, mat1, mat2, par_u, par_w, par_mv,
+    def get_fargs(self, mat, par_u, par_w, par_mv,
                   mode=None, term_mode=None, diff_var=None, **kwargs):
         vg, _ = self.get_mapping(par_u)
 
@@ -335,10 +335,13 @@ class SDDivGradTerm(Term):
         div_mv = self.get(par_mv, 'div')
         grad_mv = grad_as_vector(self.get(par_mv, 'grad'))
 
-        return (grad_u, grad_w, div_mv, grad_mv, mat1 * mat2, vg,
+        if mat is None:
+            mat = nm.ones((1, div_mv.shape[1], 1, 1), dtype=nm.float64)
+
+        return (grad_u, grad_w, div_mv, grad_mv, mat, vg,
                 get_default(term_mode, 1))
 
-    def get_eval_shape(self, mat1, mat2, par_u, par_w, par_mv,
+    def get_eval_shape(self, mat, par_u, par_w, par_mv,
                        mode=None, term_mode=None, diff_var=None, **kwargs):
         n_el, n_qp, dim, n_en, n_c = self.get_data_shape(par_u)
         return (n_el, 1, 1, 1), par_u.dtype
@@ -353,18 +356,18 @@ class SDConvectTerm(Term):
     :Definition:
 
     .. math::
-        \int_{\Omega_D} [ u_k \pdiff{u_i}{x_k} w_i (\nabla \cdot \Vcal)
+        \int_{\Omega} [ u_k \pdiff{u_i}{x_k} w_i (\nabla \cdot \Vcal)
         - u_k \pdiff{\Vcal_j}{x_k} \pdiff{u_i}{x_j} w_i ]
 
     :Arguments:
         - parameter_u : :math:`\ul{u}`
         - parameter_w : :math:`\ul{w}`
-        - parameter_mesh_velocity : :math:`\ul{\Vcal}`
+        - parameter_mv : :math:`\ul{\Vcal}`
     """
-    name = 'd_sd_convect'
-    arg_types = ('parameter_u', 'parameter_w', 'parameter_mesh_velocity')
+    name = 'ev_sd_convect'
+    arg_types = ('parameter_u', 'parameter_w', 'parameter_mv')
     arg_shapes = {'parameter_u' : 'D', 'parameter_w' : 'D',
-                  'parameter_mesh_velocity' : 'D'}
+                  'parameter_mv' : 'D'}
 
     function = staticmethod(terms.d_sd_convect)
 
@@ -421,7 +424,7 @@ class NSOFSurfMinDPressTerm(Term):
         - material_2 : :math:`bpress` (given pressure)
         - parameter  : :math:`p`
     """
-    name = 'd_of_ns_surf_min_d_press'
+    name = 'ev_of_ns_surf_min_d_press'
     arg_types = ('material_1', 'material_2', 'parameter')
     arg_shapes = {'material_1' : 1, 'material_2' : 1,
                   'parameter' : 1}
@@ -474,7 +477,7 @@ class SDGradDivStabilizationTerm(Term):
     :Definition:
 
     .. math::
-        \gamma \int_{\Omega_D} [ (\nabla \cdot \ul{u}) (\nabla \cdot \ul{w})
+        \gamma \int_{\Omega} [ (\nabla \cdot \ul{u}) (\nabla \cdot \ul{w})
         (\nabla \cdot \ul{\Vcal})
         - \pdiff{u_i}{x_k} \pdiff{\Vcal_k}{x_i} (\nabla \cdot \ul{w})
         - (\nabla \cdot \ul{u}) \pdiff{w_i}{x_k} \pdiff{\Vcal_k}{x_i} ]
@@ -483,15 +486,15 @@ class SDGradDivStabilizationTerm(Term):
         - material    : :math:`\gamma`
         - parameter_u : :math:`\ul{u}`
         - parameter_w : :math:`\ul{w}`
-        - parameter_mesh_velocity : :math:`\ul{\Vcal}`
+        - parameter_mv : :math:`\ul{\Vcal}`
         - mode        : 1 (sensitivity) or 0 (original term value)
     """
-    name = 'd_sd_st_grad_div'
+    name = 'ev_sd_st_grad_div'
     arg_types = ('material', 'parameter_u', 'parameter_w',
-                 'parameter_mesh_velocity')
+                 'parameter_mv')
     arg_shapes = {'material' : '1, 1',
                   'parameter_u' : 'D', 'parameter_w' : 'D',
-                  'parameter_mesh_velocity' : 'D'}
+                  'parameter_mv' : 'D'}
 
     function = staticmethod(terms.d_sd_st_grad_div)
 
@@ -532,15 +535,15 @@ class SDSUPGCStabilizationTerm(Term):
         - parameter_b : :math:`\ul{b}`
         - parameter_u : :math:`\ul{u}`
         - parameter_w : :math:`\ul{w}`
-        - parameter_mesh_velocity : :math:`\ul{\Vcal}`
+        - parameter_mv : :math:`\ul{\Vcal}`
         - mode        : 1 (sensitivity) or 0 (original term value)
     """
-    name = 'd_sd_st_supg_c'
+    name = 'ev_sd_st_supg_c'
     arg_types = ('material', 'parameter_b', 'parameter_u', 'parameter_w',
-                'parameter_mesh_velocity')
+                'parameter_mv')
     arg_shapes = {'material' : '1, 1',
                   'parameter_b' : 'D', 'parameter_u' : 'D', 'parameter_w' : 'D',
-                  'parameter_mesh_velocity' : 'D'}
+                  'parameter_mv' : 'D'}
 
     function = staticmethod(terms.d_sd_st_supg_c)
 
@@ -580,15 +583,15 @@ class SDPSPGCStabilizationTerm(Term):
         - parameter_b : :math:`\ul{b}`
         - parameter_u : :math:`\ul{u}`
         - parameter_r : :math:`r`
-        - parameter_mesh_velocity : :math:`\ul{\Vcal}`
+        - parameter_mv : :math:`\ul{\Vcal}`
         - mode        : 1 (sensitivity) or 0 (original term value)
     """
-    name = 'd_sd_st_pspg_c'
+    name = 'ev_sd_st_pspg_c'
     arg_types = ('material', 'parameter_b', 'parameter_u', 'parameter_r',
-                'parameter_mesh_velocity')
+                'parameter_mv')
     arg_shapes = {'material' : '1, 1',
                   'parameter_b' : 'D', 'parameter_u' : 'D', 'parameter_r' : 1,
-                  'parameter_mesh_velocity' : 'D'}
+                  'parameter_mv' : 'D'}
 
     function = staticmethod(terms.d_sd_st_pspg_c)
 
@@ -625,15 +628,15 @@ class SDPSPGPStabilizationTerm(Term):
         - material    : :math:`\tau_K`
         - parameter_r : :math:`r`
         - parameter_p : :math:`p`
-        - parameter_mesh_velocity : :math:`\ul{\Vcal}`
+        - parameter_mv : :math:`\ul{\Vcal}`
         - mode        : 1 (sensitivity) or 0 (original term value)
     """
-    name = 'd_sd_st_pspg_p'
+    name = 'ev_sd_st_pspg_p'
     arg_types = ('material', 'parameter_r', 'parameter_p',
-                'parameter_mesh_velocity')
+                'parameter_mv')
     arg_shapes = {'material' : '1, 1',
                   'parameter_r' : 1, 'parameter_p' : 1,
-                  'parameter_mesh_velocity' : 'D'}
+                  'parameter_mv' : 'D'}
 
     function = staticmethod(terms.d_sd_st_pspg_p)
 
