@@ -1,27 +1,33 @@
 import numpy as nm
 
-from sfepy.base.base import assert_
 from sfepy.linalg import dot_sequences
 from sfepy.terms.terms import Term, terms
 from sfepy.terms.terms_th import THTerm, ETHTerm
 
-class DotProductVolumeTerm(Term):
+
+class DotProductTerm(Term):
     r"""
-    Volume :math:`L^2(\Omega)` weighted dot product for both scalar and vector
-    fields. Can be evaluated. Can use derivatives.
+    Volume and surface :math:`L^2()` weighted dot product for both
+    scalar and vector fields. If the region is a surface and
+    either virtual or state variable is a vector, the orientation
+    of the normal vectors is outwards to the parent region of the virtual
+    variable. Can be evaluated. Can use derivatives.
 
     :Definition:
 
     .. math::
-        \int_\Omega q p \mbox{ , } \int_\Omega \ul{v} \cdot \ul{u}
+        \int_{\cal{D}} q p \mbox{ , } \int_{\cal{D}} \ul{v} \cdot \ul{u}
         \mbox{ , }
-        \int_\Omega p r \mbox{ , } \int_\Omega \ul{u} \cdot \ul{w} \\
-        \int_\Omega c q p \mbox{ , } \int_\Omega c \ul{v} \cdot \ul{u}
+        \int_\Gamma \ul{v} \cdot \ul{n} p \mbox{ , }
+        \int_\Gamma q \ul{n} \cdot \ul{u} \mbox{ , }
+        \int_{\cal{D}} p r \mbox{ , } \int_{\cal{D}} \ul{u} \cdot \ul{w}
+        \mbox{ , } \int_\Gamma \ul{w} \cdot \ul{n} p \\
+        \int_{\cal{D}} c q p \mbox{ , } \int_{\cal{D}} c \ul{v} \cdot \ul{u}
         \mbox{ , }
-        \int_\Omega c p r \mbox{ , } \int_\Omega c \ul{u} \cdot \ul{w} \\
-        \int_\Omega \ul{v} \cdot \ull{M} \cdot \ul{u}
+        \int_{\cal{D}} c p r \mbox{ , } \int_{\cal{D}} c \ul{u} \cdot \ul{w} \\
+        \int_{\cal{D}} \ul{v} \cdot \ull{M} \cdot \ul{u}
         \mbox{ , }
-        \int_\Omega \ul{u} \cdot \ull{M} \cdot \ul{w}
+        \int_{\cal{D}} \ul{u} \cdot \ull{M} \cdot \ul{w}
 
     :Arguments 1:
         - material : :math:`c` or :math:`\ull{M}` (optional)
@@ -33,17 +39,33 @@ class DotProductVolumeTerm(Term):
         - parameter_1 : :math:`p` or :math:`\ul{u}`
         - parameter_2 : :math:`r` or :math:`\ul{w}`
     """
-    name = 'dw_volume_dot'
+    name = 'dw_dot'
     arg_types = (('opt_material', 'virtual', 'state'),
                  ('opt_material', 'parameter_1', 'parameter_2'))
-    arg_shapes = [{'opt_material' : '1, 1', 'virtual' : (1, 'state'),
-                   'state' : 1, 'parameter_1' : 1, 'parameter_2' : 1},
-                  {'opt_material' : None},
-                  {'opt_material' : '1, 1', 'virtual' : ('D', 'state'),
-                   'state' : 'D', 'parameter_1' : 'D', 'parameter_2' : 'D'},
-                  {'opt_material' : 'D, D'},
-                  {'opt_material' : None}]
+    arg_shapes_dict = {
+        'volume': [{'opt_material' : '1, 1', 'virtual' : (1, 'state'),
+                    'state' : 1, 'parameter_1' : 1, 'parameter_2' : 1},
+                   {'opt_material' : None},
+                   {'opt_material' : '1, 1', 'virtual' : ('D', 'state'),
+                    'state' : 'D', 'parameter_1' : 'D', 'parameter_2' : 'D'},
+                   {'opt_material' : 'D, D'},
+                   {'opt_material' : None}],
+        'surface': [{'opt_material' : '1, 1', 'virtual' : (1, 'state'),
+                     'state' : 1, 'parameter_1' : 1, 'parameter_2' : 1},
+                    {'opt_material' : None},
+                    {'opt_material' : '1, 1', 'virtual' : (1, None),
+                    'state' : 'D'},
+                    {'opt_material' : None},
+                    {'opt_material' : '1, 1', 'virtual' : ('D', None),
+                    'state' : 1},
+                    {'opt_material' : None},
+                    {'opt_material' : '1, 1', 'virtual' : ('D', 'state'),
+                    'state' : 'D', 'parameter_1' : 'D', 'parameter_2' : 'D'},
+                    {'opt_material' : 'D, D'},
+                    {'opt_material' : None}]
+    }
     modes = ('weak', 'eval')
+    integration = 'by_region'
 
     @staticmethod
     def dw_dot(out, mat, val_qp, vgeo, sgeo, fun, fmode):
@@ -86,7 +108,7 @@ class DotProductVolumeTerm(Term):
         if mode == 'weak':
             if mat is None:
                 n_cell, n_qp, dim, n_n, n_c = self.get_data_shape(state)
-                mat = nm.ones((n_cell, n_qp, 1, 1), dtype=nm.float64)
+                mat = nm.ones((1, n_qp, 1, 1), dtype=nm.float64)
 
             sgeo, _ = self.get_mapping(state)
 
@@ -139,57 +161,8 @@ class DotProductVolumeTerm(Term):
         else:
             self.function = self.d_dot
 
-class DotProductSurfaceTerm(DotProductVolumeTerm):
-    r"""
-    Surface :math:`L^2(\Gamma)` dot product for both scalar and vector
-    fields.
 
-    :Definition:
-
-    .. math::
-        \int_\Gamma q p \mbox{ , } \int_\Gamma \ul{v} \cdot \ul{u}
-        \mbox{ , }
-        \int_\Gamma \ul{v} \cdot \ul{n} p \mbox{ , }
-        \int_\Gamma q \ul{n} \cdot \ul{u} \mbox{ , }
-        \int_\Gamma p r \mbox{ , } \int_\Gamma \ul{u} \cdot \ul{w}
-        \mbox{ , } \int_\Gamma \ul{w} \cdot \ul{n} p \\
-        \int_\Gamma c q p \mbox{ , } \int_\Gamma c \ul{v} \cdot \ul{u}
-        \mbox{ , }
-        \int_\Gamma c p r \mbox{ , } \int_\Gamma c \ul{u} \cdot \ul{w} \\
-        \int_\Gamma \ul{v} \cdot \ull{M} \cdot \ul{u}
-        \mbox{ , }
-        \int_\Gamma \ul{u} \cdot \ull{M} \cdot \ul{w}
-
-    :Arguments 1:
-        - material : :math:`c` or :math:`\ull{M}` (optional)
-        - virtual  : :math:`q` or :math:`\ul{v}`
-        - state    : :math:`p` or :math:`\ul{u}`
-
-    :Arguments 2:
-        - material    : :math:`c` or :math:`\ull{M}` (optional)
-        - parameter_1 : :math:`p` or :math:`\ul{u}`
-        - parameter_2 : :math:`r` or :math:`\ul{w}`
-    """
-    name = 'dw_surface_dot'
-    arg_types = (('opt_material', 'virtual', 'state'),
-                 ('opt_material', 'parameter_1', 'parameter_2'))
-    arg_shapes = [{'opt_material' : '1, 1', 'virtual' : (1, 'state'),
-                   'state' : 1, 'parameter_1' : 1, 'parameter_2' : 1},
-                  {'opt_material' : None},
-                  {'opt_material' : '1, 1', 'virtual' : (1, None),
-                   'state' : 'D'},
-                  {'opt_material' : None},
-                  {'opt_material' : '1, 1', 'virtual' : ('D', None),
-                   'state' : 1},
-                  {'opt_material' : None},
-                  {'opt_material' : '1, 1', 'virtual' : ('D', 'state'),
-                   'state' : 'D', 'parameter_1' : 'D', 'parameter_2' : 'D'},
-                  {'opt_material' : 'D, D'},
-                  {'opt_material' : None}]
-    modes = ('weak', 'eval')
-    integration = 'surface'
-
-class BCNewtonTerm(DotProductSurfaceTerm):
+class BCNewtonTerm(DotProductTerm):
     r"""
     Newton boundary condition term.
 
@@ -209,12 +182,14 @@ class BCNewtonTerm(DotProductSurfaceTerm):
     arg_shapes = {'material_1' : '1, 1', 'material_2' : '1, 1',
                   'virtual' : (1, 'state'), 'state' : 1}
     mode = 'weak'
+    integration = 'surface'
+    arg_shapes_dict = None
 
     def get_fargs(self, alpha, p_outer, virtual, state,
                   mode=None, term_mode=None, diff_var=None, **kwargs):
-        fargs = DotProductSurfaceTerm.get_fargs(self, alpha, virtual, state,
-                                                mode, term_mode, diff_var,
-                                                **kwargs)
+        fargs = DotProductTerm.get_fargs(self, alpha, virtual, state,
+                                         mode, term_mode, diff_var,
+                                         **kwargs)
         fargs = fargs[:1] + (fargs[1] - p_outer,) + fargs[2:]
 
         return fargs
@@ -320,8 +295,8 @@ class VectorDotGradScalarTerm(Term):
         \int_{\Omega} \ul{u} \cdot \nabla q \\
         \int_{\Omega} c \ul{v} \cdot \nabla p \mbox{ , }
         \int_{\Omega} c \ul{u} \cdot \nabla q \\
-        \int_{\Omega} \ul{v} \cdot \ull{M} \cdot \nabla p \mbox{ , }
-        \int_{\Omega} \ul{u} \cdot \ull{M} \cdot \nabla q
+        \int_{\Omega} \ul{v} \cdot (\ull{M} \nabla p) \mbox{ , }
+        \int_{\Omega} \ul{u} \cdot (\ull{M} \nabla q)
 
     :Arguments 1:
         - material : :math:`c` or :math:`\ull{M}` (optional)
@@ -398,7 +373,7 @@ class VectorDotGradScalarTerm(Term):
         self.function = {
             'v_weak' : terms.dw_v_dot_grad_s_vw,
             's_weak' : terms.dw_v_dot_grad_s_sw,
-            'eval' : DotProductVolumeTerm.d_dot,
+            'eval' : DotProductTerm.d_dot,
         }[self.mode]
 
 class VectorDotScalarTerm(Term):
@@ -441,7 +416,7 @@ class VectorDotScalarTerm(Term):
     def dw_dot(out, mat, val_qp, bfve, bfsc, geo, fmode):
 
         nel, nqp, dim, nc = mat.shape
-        nen = bfve.shape[2]
+        nen = bfve.shape[3]
 
         status1 = 0
         if fmode in [0, 1, 3]:
@@ -481,11 +456,11 @@ class VectorDotScalarTerm(Term):
         coef = coef.reshape(coef.shape[:2] + (dim, 1))
 
         if mode == 'weak':
-            apv, vgv = self.get_approximation(vvar)
-            aps, vgs = self.get_approximation(svar)
+            vgv, _ = self.get_mapping(vvar)
+            vgs, _ = self.get_mapping(svar)
 
-            bfve = apv.get_base('v', 0, self.integral)
-            bfsc = aps.get_base('v', 0, self.integral)
+            bfve = vgv.bf
+            bfsc = vgs.bf
 
             if self.mode == 'v_weak':
                 qp_var, geo, fmode = svar, vgv, 0
@@ -575,13 +550,12 @@ class ScalarDotGradIScalarTerm(Term):
                 grad = nm.array([0], ndmin=4, dtype=nm.float64)
                 fmode = 1
 
-            ap, vg = self.get_approximation(virtual)
-            aps, vgs = self.get_approximation(state)
+            vg, _ = self.get_mapping(virtual)
+            vgs, _ = self.get_mapping(state)
 
-            bf = aps.get_base('v', 0, self.integral)
             idx = int(material)
 
-            return bf, vg, grad, idx, fmode
+            return vgs.bf, vg, grad, idx, fmode
 
         else:
             raise ValueError('unsupported evaluation mode in %s! (%s)'
@@ -589,3 +563,72 @@ class ScalarDotGradIScalarTerm(Term):
 
     def set_arg_types(self):
         self.function = self.dw_fun
+
+class ScalarDotMGradScalarTerm(Term):
+    r"""
+    Volume dot product of a scalar gradient dotted with a material vector with
+    a scalar.
+
+    :Definition:
+
+    .. math::
+        \int_{\Omega} q \ul{y} \cdot \nabla p \mbox{ , }
+        \int_{\Omega} p \ul{y} \cdot \nabla q
+
+    :Arguments 1:
+        - material : :math:`\ul{y}`
+        - virtual  : :math:`q`
+        - state    : :math:`p`
+
+    :Arguments 2:
+        - material : :math:`\ul{y}`
+        - state    : :math:`p`
+        - virtual  : :math:`q`
+    """
+    name = 'dw_s_dot_mgrad_s'
+    arg_types = (('material', 'virtual', 'state'),
+                 ('material', 'state', 'virtual'))
+    arg_shapes = [{'material' : 'D, 1',
+                   'virtual/grad_state' : (1, None),
+                   'state/grad_state' : 1,
+                   'virtual/grad_virtual' : (1, None),
+                   'state/grad_virtual' : 1}]
+    modes = ('grad_state', 'grad_virtual')
+
+    @staticmethod
+    def function(out, out_qp, geo, fmode):
+        status = geo.integrate(out, out_qp)
+        return status
+
+    def get_fargs(self, mat, var1, var2,
+                  mode=None, term_mode=None, diff_var=None, **kwargs):
+        vg1, _ = self.get_mapping(var1)
+        vg2, _ = self.get_mapping(var2)
+
+        if diff_var is None:
+            if self.mode == 'grad_state':
+                geo = vg1
+                bf_t = vg1.bf.transpose((0, 1, 3, 2))
+                val_qp = self.get(var2, 'grad')
+                out_qp = bf_t * dot_sequences(mat, val_qp, 'ATB')
+
+            else:
+                geo = vg2
+                val_qp = self.get(var1, 'val')
+                out_qp = dot_sequences(vg2.bfg, mat, 'ATB') * val_qp
+
+            fmode = 0
+
+        else:
+            if self.mode == 'grad_state':
+                geo = vg1
+                bf_t = vg1.bf.transpose((0, 1, 3, 2))
+                out_qp = bf_t * dot_sequences(mat, vg2.bfg, 'ATB')
+
+            else:
+                geo = vg2
+                out_qp = dot_sequences(vg2.bfg, mat, 'ATB') * vg1.bf
+
+            fmode = 1
+
+        return out_qp, geo, fmode
