@@ -1,6 +1,5 @@
 import numpy as nm
 
-from sfepy.base.base import use_method_with_name, assert_
 from sfepy.linalg import dot_sequences
 from sfepy.homogenization.utils import iter_sym
 from sfepy.terms.terms import Term, terms
@@ -124,7 +123,7 @@ class LinearElasticIsotropicTerm(LinearElasticTerm):
                   mode=None, term_mode=None, diff_var=None, **kwargs):
         from sfepy.mechanics.matcoefs import stiffness_from_lame
 
-        mat = stiffness_from_lame(self.region.dim, lam, mu)
+        mat = stiffness_from_lame(self.region.dim, lam, mu)[:, :, 0, 0, :, :]
         return LinearElasticTerm.get_fargs(self, mat, virtual, state,
                                            mode=mode, term_mode=term_mode,
                                            diff_var=diff_var, **kwargs)
@@ -153,7 +152,7 @@ class SDLinearElasticTerm(Term):
         - parameter_u : :math:`\ul{u}`
         - parameter_mesh_velocity : :math:`\ul{\Vcal}`
     """
-    name = 'd_sd_lin_elastic'
+    name = 'ev_sd_lin_elastic'
     arg_types = ('material', 'parameter_w', 'parameter_u',
                  'parameter_mesh_velocity')
     arg_shapes = {'material' : 'S, S',
@@ -276,11 +275,11 @@ class LinearElasticETHTerm(ETHTerm):
 
 class LinearPrestressTerm(Term):
     r"""
-    Linear prestress term, with the prestress :math:`\sigma_{ij}` given in
-    the usual vector form exploiting symmetry: in 3D it has 6 components
+    Linear prestress term, with the prestress :math:`\sigma_{ij}` given either
+    in the usual vector form exploiting symmetry: in 3D it has 6 components
     with the indices ordered as :math:`[11, 22, 33, 12, 13, 23]`, in 2D it has
-    3 components with the indices ordered as :math:`[11, 22, 12]`. Can be
-    evaluated.
+    3 components with the indices ordered as :math:`[11, 22, 12]`, or in the
+    matrix (possibly non-symmetric) form. Can be evaluated.
 
     :Definition:
 
@@ -298,19 +297,31 @@ class LinearPrestressTerm(Term):
     name = 'dw_lin_prestress'
     arg_types = (('material', 'virtual'),
                  ('material', 'parameter'))
-    arg_shapes = {'material' : 'S, 1', 'virtual' : ('D', None),
-                  'parameter' : 'D'}
+    arg_shapes = [{'material' : 'S, 1', 'virtual' : ('D', None),
+                  'parameter' : 'D'},
+                  {'material' : 'D, D'}]
     modes = ('weak', 'eval')
 
     def get_fargs(self, mat, virtual,
                   mode=None, term_mode=None, diff_var=None, **kwargs):
         vg, _ = self.get_mapping(virtual)
 
+        sh = mat.shape
+        is_nonsym = sh[2] == sh[3] == vg.dim and not(vg.dim == 1)
+
+        if is_nonsym:
+            mat = mat.reshape(sh[:2] + (vg.dim**2, 1))
+
         if mode == 'weak':
             return mat, vg
 
         else:
-            strain = self.get(virtual, 'cauchy_strain')
+            if is_nonsym:
+                strain = self.get(virtual, 'grad').transpose((0,1,3,2))
+                nel, nqp, nr, nc = strain.shape
+                strain = strain.reshape((nel, nqp, nr*nc, 1))
+            else:
+                strain = self.get(virtual, 'cauchy_strain')
 
             fmode = {'eval' : 0, 'el_avg' : 1, 'qp' : 2}.get(mode, 1)
             return strain, mat, vg, fmode
@@ -389,7 +400,7 @@ class CauchyStrainTerm(Term):
     :Definition:
 
     .. math::
-        \int_{\Omega} \ull{e}(\ul{w})
+        \int_{\cal{D}} \ull{e}(\ul{w})
 
     .. math::
         \mbox{vector for } K \from \Ical_h: \int_{T_K} \ull{e}(\ul{w}) /
@@ -404,6 +415,8 @@ class CauchyStrainTerm(Term):
     name = 'ev_cauchy_strain'
     arg_types = ('parameter',)
     arg_shapes = {'parameter' : 'D'}
+    integration = 'by_region'
+    surface_integration = 'surface_extra'
 
     @staticmethod
     def function(out, strain, vg, fmode):
@@ -433,34 +446,8 @@ class CauchyStrainTerm(Term):
         if mode != 'qp':
             n_qp = 1
 
-        return (n_el, n_qp, dim * (dim + 1) / 2, 1), parameter.dtype
+        return (n_el, n_qp, dim * (dim + 1) // 2, 1), parameter.dtype
 
-class CauchyStrainSTerm(CauchyStrainTerm):
-    r"""
-    Evaluate Cauchy strain tensor on a surface region.
-
-    See :class:`CauchyStrainTerm`.
-
-    Supports 'eval', 'el_avg' and 'qp' evaluation modes.
-
-    :Definition:
-
-    .. math::
-        \int_{\Gamma} \ull{e}(\ul{w})
-
-    .. math::
-        \mbox{vector for } K \from \Ical_h: \int_{T_K} \ull{e}(\ul{w}) /
-        \int_{T_K} 1
-
-    .. math::
-        \ull{e}(\ul{w})|_{qp}
-
-    :Arguments:
-        - parameter : :math:`\ul{w}`
-    """
-    name = 'ev_cauchy_strain_s'
-    arg_types = ('parameter',)
-    integration = 'surface_extra'
 
 class CauchyStressTerm(Term):
     r"""
@@ -475,7 +462,7 @@ class CauchyStressTerm(Term):
     :Definition:
 
     .. math::
-        \int_{\Omega} D_{ijkl} e_{kl}(\ul{w})
+        \int_{\cal{D}} D_{ijkl} e_{kl}(\ul{w})
 
     .. math::
         \mbox{vector for } K \from \Ical_h:
@@ -491,6 +478,8 @@ class CauchyStressTerm(Term):
     name = 'ev_cauchy_stress'
     arg_types = ('material', 'parameter')
     arg_shapes = {'material' : 'S, S', 'parameter' : 'D'}
+    integration = 'by_region'
+    surface_integration = 'surface_extra'
 
     @staticmethod
     def function(out, coef, strain, mat, vg, fmode):
@@ -523,7 +512,7 @@ class CauchyStressTerm(Term):
         if mode != 'qp':
             n_qp = 1
 
-        return (n_el, n_qp, dim * (dim + 1) / 2, 1), parameter.dtype
+        return (n_el, n_qp, dim * (dim + 1) // 2, 1), parameter.dtype
 
 class CauchyStressTHTerm(CauchyStressTerm, THTerm):
     r"""
@@ -636,3 +625,265 @@ class CauchyStressETHTerm(CauchyStressTerm, ETHTerm):
         out = CauchyStressTerm.get_eval_shape(self, mat0, parameter, mode,
                                               term_mode, diff_var, **kwargs)
         return out
+
+class NonsymElasticTerm(Term):
+    r"""
+    Elasticity term with non-symmetric gradient. The indices of matrix
+    :math:`D_{ijkl}` are ordered as
+    :math:`[11, 12, 13, 21, 22, 23, 31, 32, 33]` in 3D and as
+    :math:`[11, 12, 21, 22]` in 2D.
+
+    :Definition:
+
+    .. math::
+        \int_{\Omega} \ull{D} \nabla\ul{u} : \nabla\ul{v}
+
+    :Arguments 1:
+        - material : :math:`\ull{D}`
+        - virtual  : :math:`\ul{v}`
+        - state    : :math:`\ul{u}`
+
+    :Arguments 2:
+        - material    : :math:`\ull{D}`
+        - parameter_1 : :math:`\ul{w}`
+        - parameter_2 : :math:`\ul{u}`
+    """
+    name = 'dw_nonsym_elastic'
+    arg_types = (('material', 'virtual', 'state'),
+                 ('material', 'parameter_1', 'parameter_2'))
+    arg_shapes = {'material' : 'D2, D2', 'virtual' : ('D', 'state'),
+                  'state' : 'D', 'parameter_1' : 'D', 'parameter_2' : 'D'}
+    modes = ('weak', 'eval')
+    geometries = ['2_3', '2_4', '3_4', '3_8']
+
+    def get_fargs(self, mat, virtual, state,
+                  mode=None, term_mode=None, diff_var=None, **kwargs):
+        vg, _ = self.get_mapping(state)
+
+        if mode == 'weak':
+            if diff_var is None:
+                grad = self.get(state, 'grad').transpose((0,1,3,2))
+                nel, nqp, nr, nc = grad.shape
+                grad = grad.reshape((nel,nqp,nr*nc,1))
+                fmode = 0
+
+            else:
+                grad = nm.array([0], ndmin=4, dtype=nm.float64)
+                fmode = 1
+
+            return grad, mat, vg, fmode
+
+        elif mode == 'eval':
+            grad1 = self.get(virtual, 'grad').transpose((0,1,3,2))
+            grad2 = self.get(state, 'grad').transpose((0,1,3,2))
+            nel, nqp, nr, nc = grad1.shape
+
+            return 1.0,\
+                   grad1.reshape((nel,nqp,nr*nc,1)),\
+                   grad2.reshape((nel,nqp,nr*nc,1)),\
+                   mat, vg
+
+        else:
+            raise ValueError('unsupported evaluation mode in %s! (%s)'
+                             % (self.name, mode))
+
+    def get_eval_shape(self, mat, virtual, state,
+                       mode=None, term_mode=None, diff_var=None, **kwargs):
+        n_el, n_qp, dim, n_en, n_c = self.get_data_shape(state)
+
+        return (n_el, 1, 1, 1), state.dtype
+
+    def set_arg_types(self):
+        if self.mode == 'weak':
+            self.function = terms.dw_nonsym_elastic
+
+        else:
+            self.function = terms.d_lin_elastic
+
+def _build_wave_strain_op(vec, bf):
+    dim = len(vec)
+
+    if dim == 2:
+        n0, n1 = vec
+        nmat = nm.array([[n0, 0],
+                         [0, n1],
+                         [n1, n0]], dtype=nm.float64)
+
+    else:
+        n0, n1, n2 = vec
+        nmat = nm.array([[n0, 0, 0],
+                         [0, n1, 0],
+                         [0, 0, n2],
+                         [n1, n0, 0],
+                         [n2, 0, n0],
+                         [0, n2, n1]], dtype=nm.float64)
+
+    out = nm.einsum('ik,cqkj->cqij', nmat, bf)
+    return out
+
+from sfepy.base.compat import block
+
+def _build_cauchy_strain_op(bfg):
+    dim = bfg.shape[2]
+    if dim == 2:
+        g1, g2 = bfg[..., 0:1, :], bfg[..., 1:2, :]
+        zz = nm.zeros_like(g1)
+        out = block([[g1, zz],
+                     [zz, g2],
+                     [g2, g1]])
+
+    else:
+        g1, g2, g3 = bfg[..., 0:1, :], bfg[..., 1:2, :], bfg[..., 2:3, :]
+        zz = nm.zeros_like(g1)
+        out = block([[g1, zz, zz],
+                     [zz, g2, zz],
+                     [zz, zz, g3],
+                     [g2, g1, zz],
+                     [g3, zz, g1],
+                     [zz, g3, g2]])
+
+    return out
+
+class ElasticWaveTerm(Term):
+    r"""
+    Elastic dispersion term involving the wave strain :math:`g_{ij}`,
+    :math:`g_{ij}(\ul{u}) = \frac{1}{2}(u_i \kappa_j + \kappa_i u_j)`, with the
+    wave vector :math:`\ul{\kappa}`. :math:`D_{ijkl}` is given in the usual
+    matrix form exploiting symmetry: in 3D it is :math:`6\times6` with the
+    indices ordered as :math:`[11, 22, 33, 12, 13, 23]`, in 2D it is
+    :math:`3\times3` with the indices ordered as :math:`[11, 22, 12]`.
+
+    :Definition:
+
+    .. math::
+        \int_{\Omega} D_{ijkl}\ g_{ij}(\ul{v}) g_{kl}(\ul{u})
+
+    :Arguments:
+        - material_1 : :math:`D_{ijkl}`
+        - material_2 : :math:`\ul{\kappa}`
+        - virtual    : :math:`\ul{v}`
+        - state      : :math:`\ul{u}`
+    """
+    name = 'dw_elastic_wave'
+    arg_types = ('material_1', 'material_2', 'virtual', 'state')
+    arg_shapes = {'material_1' : 'S, S', 'material_2' : '.: D',
+                  'virtual' : ('D', 'state'), 'state' : 'D'}
+    geometries = ['2_3', '2_4', '3_4', '3_8']
+
+    @staticmethod
+    def function(out, out_qp, geo, fmode):
+        status = geo.integrate(out, out_qp)
+        return status
+
+    def get_fargs(self, mat, kappa, virtual, state,
+                  mode=None, term_mode=None, diff_var=None, **kwargs):
+        from sfepy.discrete.variables import create_adof_conn, expand_basis
+
+        geo, _ = self.get_mapping(state)
+
+        n_el, n_qp, dim, n_en, n_c = self.get_data_shape(virtual)
+
+        ebf = expand_basis(geo.bf, dim)
+
+        mat = Term.tile_mat(mat, n_el)
+        gmat = _build_wave_strain_op(kappa, ebf)
+
+        if diff_var is None:
+            econn = state.field.get_econn('volume', self.region)
+            adc = create_adof_conn(nm.arange(state.n_dof, dtype=nm.int32),
+                                   econn, n_c, 0)
+            vals = state()[adc]
+            # Same as nm.einsum('qij,cj->cqi', gmat[0], vals)[..., None]
+            aux = dot_sequences(gmat, vals[:, None, :, None])
+            out_qp = dot_sequences(gmat, dot_sequences(mat, aux), 'ATB')
+            fmode = 0
+
+        else:
+            out_qp = dot_sequences(gmat, dot_sequences(mat, gmat), 'ATB')
+            fmode = 1
+
+        return out_qp, geo, fmode
+
+class ElasticWaveCauchyTerm(Term):
+    r"""
+    Elastic dispersion term involving the wave strain :math:`g_{ij}`,
+    :math:`g_{ij}(\ul{u}) = \frac{1}{2}(u_i \kappa_j + \kappa_i u_j)`, with the
+    wave vector :math:`\ul{\kappa}` and the elastic strain :math:`e_{ij}`.
+    :math:`D_{ijkl}` is given in the usual matrix form exploiting symmetry: in
+    3D it is :math:`6\times6` with the indices ordered as :math:`[11, 22, 33,
+    12, 13, 23]`, in 2D it is :math:`3\times3` with the indices ordered as
+    :math:`[11, 22, 12]`.
+
+    :Definition:
+
+    .. math::
+        \int_{\Omega} D_{ijkl}\ g_{ij}(\ul{v}) e_{kl}(\ul{u}) \;,
+        \int_{\Omega} D_{ijkl}\ g_{ij}(\ul{u}) e_{kl}(\ul{v})
+
+    :Arguments 1:
+        - material_1 : :math:`D_{ijkl}`
+        - material_2 : :math:`\ul{\kappa}`
+        - virtual    : :math:`\ul{v}`
+        - state      : :math:`\ul{u}`
+
+    :Arguments 2:
+        - material_1 : :math:`D_{ijkl}`
+        - material_2 : :math:`\ul{\kappa}`
+        - state      : :math:`\ul{u}`
+        - virtual    : :math:`\ul{v}`
+    """
+    name = 'dw_elastic_wave_cauchy'
+    arg_types = (('material_1', 'material_2', 'virtual', 'state'),
+                 ('material_1', 'material_2', 'state', 'virtual'))
+    arg_shapes = {'material_1' : 'S, S', 'material_2' : '.: D',
+                  'virtual' : ('D', 'state'), 'state' : 'D'}
+    geometries = ['2_3', '2_4', '3_4', '3_8']
+    modes = ('ge', 'eg')
+
+    @staticmethod
+    def function(out, out_qp, geo, fmode):
+        status = geo.integrate(out, out_qp)
+        return status
+
+    def get_fargs(self, mat, kappa, gvar, evar,
+                  mode=None, term_mode=None, diff_var=None, **kwargs):
+        from sfepy.discrete.variables import create_adof_conn, expand_basis
+
+        geo, _ = self.get_mapping(evar)
+
+        n_el, n_qp, dim, n_en, n_c = self.get_data_shape(gvar)
+
+        ebf = expand_basis(geo.bf, dim)
+
+        mat = Term.tile_mat(mat, n_el)
+        gmat = _build_wave_strain_op(kappa, ebf)
+        emat = _build_cauchy_strain_op(geo.bfg)
+
+        if diff_var is None:
+            avar = evar if self.mode == 'ge' else gvar
+            econn = avar.field.get_econn('volume', self.region)
+            adc = create_adof_conn(nm.arange(avar.n_dof, dtype=nm.int32),
+                                   econn, n_c, 0)
+            vals = avar()[adc]
+
+            if self.mode == 'ge':
+                # Same as aux = self.get(avar, 'cauchy_strain'),
+                aux = dot_sequences(emat, vals[:, None, :, None])
+                out_qp = dot_sequences(gmat, dot_sequences(mat, aux), 'ATB')
+
+            else:
+                aux = dot_sequences(gmat, vals[:, None, :, None])
+                out_qp = dot_sequences(emat, dot_sequences(mat, aux), 'ATB')
+
+            fmode = 0
+
+        else:
+            if self.mode == 'ge':
+                out_qp = dot_sequences(gmat, dot_sequences(mat, emat), 'ATB')
+
+            else:
+                out_qp = dot_sequences(emat, dot_sequences(mat, gmat), 'ATB')
+
+            fmode = 1
+
+        return out_qp, geo, fmode
