@@ -1,6 +1,6 @@
 import numpy as nm
 
-from sfepy.linalg import dot_sequences
+from sfepy.linalg import dot_sequences, insert_strided_axis
 from sfepy.terms.terms import Term, terms
 
 class DivGradTerm(Term):
@@ -175,6 +175,56 @@ class LinearConvectTerm(Term):
             raise ValueError('unsupported evaluation mode in %s! (%s)'
                              % (self.name, mode))
 
+class LinearConvect2Term(Term):
+    r"""
+    Linearized convective term with the convection velocity given as a material
+    parameter.
+
+    :Definition:
+
+    .. math::
+        \int_{\Omega} ((\ul{b} \cdot \nabla) \ul{u}) \cdot \ul{v}
+
+    .. math::
+        ((\ul{b} \cdot \nabla) \ul{u})|_{qp}
+
+    :Arguments:
+        - material : :math:`\ul{b}`
+        - virtual  : :math:`\ul{v}`
+        - state    : :math:`\ul{u}`
+    """
+    name = 'dw_lin_convect2'
+    arg_types = ('material', 'virtual', 'state')
+    arg_shapes = {'material' : 'D, 1',
+                  'virtual' : ('D', 'state'), 'state' : 'D'}
+
+    function = staticmethod(terms.dw_lin_convect)
+
+    def get_fargs(self, material, virtual, state,
+                  mode=None, term_mode=None, diff_var=None, **kwargs):
+        vg, _ = self.get_mapping(state)
+
+        if mode == 'weak':
+            if diff_var is None:
+                grad = self.get(state, 'grad').transpose((0, 1, 3, 2)).copy()
+                fmode = 0
+
+            else:
+                grad = nm.array([0], ndmin=4, dtype=nm.float64)
+                fmode = 1
+
+            return grad, material, vg, fmode
+
+        elif mode == 'qp':
+            grad = self.get(state, 'grad').transpose((0, 1, 3, 2)).copy()
+            fmode = 2
+
+            return grad, material, vg, fmode
+
+        else:
+            raise ValueError('unsupported evaluation mode in %s! (%s)'
+                             % (self.name, mode))
+
 class StokesTerm(Term):
     r"""
     Stokes problem coupling term. Corresponds to weak forms of gradient and
@@ -283,7 +333,7 @@ class GradTerm(Term):
     :Definition:
 
     .. math::
-        \int_{\Omega} \nabla p \mbox{ or } \int_{\Omega} \nabla \ul{w}
+        \int_{\cal{D}} \nabla p \mbox{ or } \int_{\cal{D}} \nabla \ul{w}
 
     .. math::
         \mbox{vector for } K \from \Ical_h: \int_{T_K} \nabla p /
@@ -299,6 +349,8 @@ class GradTerm(Term):
     name = 'ev_grad'
     arg_types = ('parameter',)
     arg_shapes = {'parameter' : 'N'}
+    integration = 'by_region'
+    surface_integration = 'surface_extra'
 
     @staticmethod
     def function(out, grad, vg, fmode):
@@ -315,8 +367,7 @@ class GradTerm(Term):
                   mode=None, term_mode=None, diff_var=None, **kwargs):
         vg, _ = self.get_mapping(parameter)
 
-        grad = self.get(parameter, 'grad')
-        grad = nm.ascontiguousarray(grad.transpose((0, 1, 3, 2)))
+        grad = self.get(parameter, 'grad', integration=self.integration)
 
         fmode = {'eval' : 0, 'el_avg' : 1, 'qp' : 2}.get(mode, 1)
 
@@ -329,7 +380,7 @@ class GradTerm(Term):
         if mode != 'qp':
             n_qp = 1
 
-        return (n_el, n_qp, n_c, dim), parameter.dtype
+        return (n_el, n_qp, dim, n_c), parameter.dtype
 
 class DivTerm(Term):
     r"""
@@ -340,7 +391,7 @@ class DivTerm(Term):
     :Definition:
 
     .. math::
-         \int_{\Omega} \nabla \cdot \ul{u}
+         \int_{\cal{D}} \nabla \cdot \ul{u}
 
     .. math::
          \mbox{vector for } K \from \Ical_h:
@@ -355,6 +406,8 @@ class DivTerm(Term):
     name = 'ev_div'
     arg_types = ('parameter',)
     arg_shapes = {'parameter' : 'D'}
+    integration = 'by_region'
+    surface_integration = 'surface_extra'
 
     @staticmethod
     def function(out, div, vg, fmode):
@@ -371,7 +424,7 @@ class DivTerm(Term):
                   mode=None, term_mode=None, diff_var=None, **kwargs):
         vg, _ = self.get_mapping(parameter)
 
-        div = self.get(parameter, 'div')
+        div = self.get(parameter, 'div', integration=self.integration)
 
         fmode = {'eval' : 0, 'el_avg' : 1, 'qp' : 2}.get(mode, 1)
 
@@ -426,6 +479,139 @@ class DivOperatorTerm(Term):
 
         return mat, vg
 
+class StokesWaveTerm(Term):
+    r"""
+    Stokes dispersion term with the wave vector :math:`\ul{\kappa}`.
+
+    :Definition:
+
+    .. math::
+        \int_{\Omega} (\ul{\kappa} \cdot \ul{v}) (\ul{\kappa} \cdot \ul{u})
+
+    :Arguments:
+        - material : :math:`\ul{\kappa}`
+        - virtual  : :math:`\ul{v}`
+        - statee   : :math:`\ul{u}`
+    """
+    name = 'dw_stokes_wave'
+    arg_types = ('material', 'virtual', 'state')
+    arg_shapes = {'material' : '.: D',
+                  'virtual' : ('D', 'state'), 'state' : 'D'}
+    geometries = ['2_3', '2_4', '3_4', '3_8']
+
+    @staticmethod
+    def function(out, out_qp, geo, fmode):
+        status = geo.integrate(out, out_qp)
+        return status
+
+    def get_fargs(self, kappa, virtual, state,
+                  mode=None, term_mode=None, diff_var=None, **kwargs):
+        from sfepy.discrete.variables import create_adof_conn, expand_basis
+
+        geo, _ = self.get_mapping(state)
+
+        n_el, n_qp, dim, n_en, n_c = self.get_data_shape(virtual)
+
+        ebf = expand_basis(geo.bf, dim)
+
+        aux = nm.einsum('i,...ij->...j', kappa, ebf)[0, :, None, :]
+        kebf = insert_strided_axis(aux, 0, n_el)
+
+        if diff_var is None:
+            econn = state.field.get_econn('volume', self.region)
+            adc = create_adof_conn(nm.arange(state.n_dof, dtype=nm.int32),
+                                   econn, n_c, 0)
+            vals = state()[adc]
+            aux = dot_sequences(kebf, vals[:, None, :, None])
+            out_qp = dot_sequences(kebf, aux, 'ATB')
+            fmode = 0
+
+        else:
+            out_qp = dot_sequences(kebf, kebf, 'ATB')
+            fmode = 1
+
+        return out_qp, geo, fmode
+
+class StokesWaveDivTerm(Term):
+    r"""
+    Stokes dispersion term with the wave vector :math:`\ul{\kappa}` and the
+    divergence operator.
+
+    :Definition:
+
+    .. math::
+        \int_{\Omega} (\ul{\kappa} \cdot \ul{v}) (\nabla \cdot \ul{u}) \;,
+        \int_{\Omega} (\ul{\kappa} \cdot \ul{u}) (\nabla \cdot \ul{v})
+
+    :Arguments 1:
+        - material : :math:`\ul{\kappa}`
+        - virtual  : :math:`\ul{v}`
+        - state    : :math:`\ul{u}`
+
+    :Arguments 2:
+        - material : :math:`\ul{\kappa}`
+        - state    : :math:`\ul{u}`
+        - virtual  : :math:`\ul{v}`
+    """
+    name = 'dw_stokes_wave_div'
+    arg_types = (('material', 'virtual', 'state'),
+                 ('material', 'state', 'virtual'))
+    arg_shapes = {'material' : '.: D',
+                  'virtual' : ('D', 'state'), 'state' : 'D'}
+    geometries = ['2_3', '2_4', '3_4', '3_8']
+    modes = ('kd', 'dk')
+
+    @staticmethod
+    def function(out, out_qp, geo, fmode):
+        status = geo.integrate(out, out_qp)
+        return status
+
+    def get_fargs(self, kappa, kvar, dvar,
+                  mode=None, term_mode=None, diff_var=None, **kwargs):
+        from sfepy.discrete.variables import create_adof_conn, expand_basis
+
+        geo, _ = self.get_mapping(dvar)
+
+        n_el, n_qp, dim, n_en, n_c = self.get_data_shape(kvar)
+
+        ebf = expand_basis(geo.bf, dim)
+
+        aux = nm.einsum('i,...ij->...j', kappa, ebf)[0, :, None, :]
+        kebf = insert_strided_axis(aux, 0, n_el)
+
+        div_bf = geo.bfg
+
+        div_bf = div_bf.reshape((n_el, n_qp, 1, dim * n_en))
+        div_bf = nm.ascontiguousarray(div_bf)
+
+        if diff_var is None:
+            avar = dvar if self.mode == 'kd' else kvar
+            econn = avar.field.get_econn('volume', self.region)
+            adc = create_adof_conn(nm.arange(avar.n_dof, dtype=nm.int32),
+                                   econn, n_c, 0)
+            vals = avar()[adc]
+
+            if self.mode == 'kd':
+                aux = dot_sequences(div_bf, vals[:, None, :, None])
+                out_qp = dot_sequences(kebf, aux, 'ATB')
+
+            else:
+                aux = dot_sequences(kebf, vals[:, None, :, None])
+                out_qp = dot_sequences(div_bf, aux, 'ATB')
+
+            fmode = 0
+
+        else:
+            if self.mode == 'kd':
+                out_qp = dot_sequences(kebf, div_bf, 'ATB')
+
+            else:
+                out_qp = dot_sequences(div_bf, kebf, 'ATB')
+
+            fmode = 1
+
+        return out_qp, geo, fmode
+
 class GradDivStabilizationTerm(Term):
     r"""
     Grad-div stabilization term ( :math:`\gamma` is a global stabilization
@@ -506,11 +692,11 @@ class PSPGCStabilizationTerm(Term):
 
     def get_fargs(self, tau, virtual, parameter, state,
                   mode=None, term_mode=None, diff_var=None, **kwargs):
-        sap, svg = self.get_approximation(virtual)
-        vap, vvg = self.get_approximation(state)
+        svg, _ = self.get_mapping(virtual)
+        vvg, _ = self.get_mapping(state)
 
         val_qp = self.get(parameter, 'val')
-        conn = vap.get_connectivity(self.region, self.integration)
+        conn = state.field.get_connectivity(self.region, self.integration)
 
         if diff_var is None:
             fmode = 0
@@ -587,10 +773,10 @@ class SUPGCStabilizationTerm(Term):
 
     def get_fargs(self, delta, virtual, parameter, state,
                   mode=None, term_mode=None, diff_var=None, **kwargs):
-        ap, vg = self.get_approximation(virtual)
+        vg, _ = self.get_mapping(virtual)
 
         val_qp = self.get(parameter, 'val')
-        conn = ap.get_connectivity(self.region, self.integration)
+        conn = virtual.field.get_connectivity(self.region, self.integration)
 
         if diff_var is None:
             fmode = 0
